from enum import Enum


class CreatorEnum(str, Enum):
    SYSTEM = "SYSTEM"
    STATIC = "STATIC"
    DYNAMIC = "DYNAMIC"


class TargetRecordTypeEnum(str, Enum):
    A = "A"
    AAAA = "AAAA"
    MX = "MX"
    NAPTR = "NAPTR"
    PTR = "PTR"
    SPF = "SPF"
    SRV = "SRV"
    TXT = "TXT"


class ZoneFormatTypeEnum(str, Enum):
    FORWARD = "FORWARD"
    IPV4 = "IPV4"
    IPV6 = "IPV6"


class FailoverServerTypeEnum(str, Enum):
    GRID = "GRID"
    EXTERNAL = "EXTERNAL"


class ProtocolTypeEnum(str, Enum):
    IPV4 = "IPV4"
    IPV6 = "IPV6"


class FingerprintTypeEnum(str, Enum):
    STANDARD = "Standard"
    CUSTOM = "Custom"


class HostAddressMatchEnum(str, Enum):
    MAC_ADDRESS = "MAC_ADDRESS"
    CLIENT_IDENTIFIER = "CLIENT_IDENTIFIER"
    RESERVED = "RESERVED"


class MatchOptionEnum(str, Enum):
    MAC_ADDRESS = "MAC_ADDRESS"
    CLIENT_ID = "CLIENT_ID"
    CIRCUIT_ID = "CIRCUIT_ID"
    REMOTE_ID = "REMOTE_ID"


class IPv6AddressTypeEnum(str, Enum):
    ADDRESS = "ADDRESS"
    PREFIX = "PREFIX"
    BOTH = "BOTH"


class LeasePerClientSettingsEnum(str, Enum):
    ONE_LEASE_PER_CLIENT = "ONE_LEASE_PER_CLIENT"
    RELEASE_MATCHING_ID = "RELEASE_MATCHING_ID"
    NEVER_RELEASE = "NEVER_RELEASE"


class ServerAssociationTypeEnum(str, Enum):
    NONE = "NONE"
    MEMBER = "MEMBER"
    FAILOVER = "FAILOVER"


class ImportActionEnum(str, Enum):
    INSERT = "I"
    MERGE = "M"
    OVERRIDE = "O"
    DELETE = "D"
    INSERT_MERGE = "IM"
    INSERT_OVERRIDE = "IO"


class DhcpTypeEnum(str, Enum):
    T_ARRAY_DOMAIN = "T_ARRAY_DOMAIN"
    T_ARRAY_INT8 = "T_ARRAY_INT8"
    T_ARRAY_INT16 = "T_ARRAY_INT16"
    T_ARRAY_INT32 = "T_ARRAY_INT32"
    T_ARRAY_IP_ADDRESS = "T_ARRAY_IP_ADDRESS"
    T_ARRAY_IP_ADDRESS_PAIR = "T_ARRAY_IP_ADDRESS_PAIR"
    T_ARRAY_UINT8 = "T_ARRAY_UINT8"
    T_ARRAY_UINT16 = "T_ARRAY_UINT16"
    T_ARRAY_UINT32 = "T_ARRAY_UINT32"
    T_DOMAIN = "T_DOMAIN"
    T_FLAG = "T_FLAG"
    T_FLAG_IP_ADDRESS = "T_FLAG_IP_ADDRESS"
    T_FLAG_TEXT = "T_FLAG_TEXT"
    T_INT8 = "T_INT8"
    T_INT16 = "T_INT16"
    T_INT32 = "T_INT32"
    T_IP_ADDRESS = "T_IP_ADDRESS"
    T_STRING = "T_STRING"
    T_TEXT = "T_TEXT"
    T_UINT8 = "T_UINT8"
    T_UINT16 = "T_UINT16"
    T_UINT32 = "T_UINT32"
    T_UINT8_1_2_4_8 = "T_UINT8_1_2_4_8"



#!/usr/bin/env python3
"""
Copyright 2023 Infoblox

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may find a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import asyncio
import getpass
import sys
from typing import Literal

import click
from click_option_group import optgroup

from ibx_sdk.logger.ibx_logger import init_logger, increase_log_level
from ibx_sdk.nios.asynchronous.gift import AsyncGift
from ibx_sdk.nios.exceptions import WapiRequestException

log = init_logger(
    logfile_name="wapi.log",
    logfile_mode="a",
    console_log=True,
    level="info",
    max_size=100000,
    num_logs=1,
)
ALGORITHMS = click.Choice(["SHA-256", "SHA-384", "SHA-512"])
USAGES = click.Choice(
    ["ADMIN", "CAPTIVE_PORTAL", "SFNT_CLIENT_CERT", "IFMAP_DHCP"]
)
ALL_USAGES = click.Choice(
    [
        "ADMIN",
        "CAPTIVE_PORTAL",
        "SFNT_CLIENT_CERT",
        "IFMAP_DHCP",
        "EAP_CA",
        "TAE_CA",
    ]
)
wapi = AsyncGift()
help_text = """
NIOS SSL Certificate Tools
"""


@click.group(
    help=help_text,
    context_settings=dict(
        max_content_width=95,
        help_option_names=["-h", "--help"],
        show_default=True,
    ),
)
def cli():
    pass


@cli.command()
@optgroup.group("Required Parameters")
@optgroup.option(
    "-g", "--grid-mgr", required=True, help="Infoblox Grid Manager"
)
@optgroup.option(
    "-m", "--member", required=True, help="Member for the certificate"
)
@optgroup.option(
    "-f",
    "--filename",
    required=True,
    help="Upload filename for the certificate",
)
@optgroup.group("Optional Parameters")
@optgroup.option(
    "--certificate-usage",
    default="ADMIN",
    type=ALL_USAGES,
    help="Certificate Usage",
)
@optgroup.option(
    "-u",
    "--username",
    default="admin",
    help="Infoblox admin username",
)
@optgroup.option(
    "-w",
    "--wapi-ver",
    default="2.11",
    show_default=True,
    help="Infoblox WAPI version",
)
@optgroup.option("--debug", is_flag=True, help="enable verbose debug output")
def upload(
    grid_mgr: str,
    member: str,
    filename: str,
    username: str,
    wapi_ver: str,
    certificate_usage: str,
    debug: bool,
):
    if debug:
        increase_log_level()

    wapi.grid_mgr = grid_mgr
    wapi.wapi_ver = wapi_ver
    password = getpass.getpass(f"Enter password for [{username}]: ")

    try:
        wapi.connect(username=username, password=password)
    except WapiRequestException as err:
        log.error(err)
        sys.exit(1)
    else:
        log.info("connected to Infoblox grid manager %s", wapi.grid_mgr)
    try:
        wapi.upload_certificate(
            member=member,
            filename=filename,
            certificate_usage=certificate_usage,
        )
    except WapiRequestException as err:
        log.error(err)
        sys.exit(1)

    log.info("finished!")
    sys.exit()


@cli.command()
@optgroup.group("Required Parameters")
@optgroup.option(
    "-g", "--grid-mgr", required=True, help="Infoblox Grid Manager"
)
@optgroup.option(
    "-m", "--member", required=True, help="Member for the certificate"
)
@optgroup.group("Optional Parameters")
@optgroup.option(
    "--certificate-usage",
    default="ADMIN",
    type=ALL_USAGES,
    help="Certificate Usage",
)
@optgroup.option(
    "-u",
    "--username",
    default="admin",
    help="Infoblox admin username",
)
@optgroup.option(
    "-w",
    "--wapi-ver",
    default="2.11",
    show_default=True,
    help="Infoblox WAPI version",
)
@optgroup.option("--debug", is_flag=True, help="enable verbose debug output")
def download(
    grid_mgr: str,
    member: str,
    username: str,
    wapi_ver: str,
    certificate_usage: Literal[str],
    debug: bool,
):
    asyncio.run(
        async_download(
            grid_mgr, member, username, wapi_ver, certificate_usage, debug
        )
    )


async def async_download(
    grid_mgr: str,
    member: str,
    username: str,
    wapi_ver: str,
    certificate_usage: Literal[str],
    debug: bool,
):
    if debug:
        increase_log_level()

    wapi.grid_mgr = grid_mgr
    wapi.wapi_ver = wapi_ver
    password = getpass.getpass(f"Enter password for [{username}]: ")

    try:
        await wapi.connect(username=username, password=password)
    except WapiRequestException as err:
        log.error(err)
        sys.exit(1)
    else:
        log.info("connected to Infoblox grid manager %s", wapi.grid_mgr)
    try:
        await wapi.download_certificate(
            member=member, certificate_usage=certificate_usage
        )
    except WapiRequestException as err:
        log.error(err)
        sys.exit(1)

    log.info("finished!")
    sys.exit()


@cli.command()
@optgroup.group("Required Parameters")
@optgroup.option(
    "-g", "--grid-mgr", required=True, help="Infoblox Grid Manager"
)
@optgroup.option(
    "-n", "--common-name", required=True, help="Common Name for the certificate"
)
@optgroup.option(
    "-m", "--member", required=True, help="Member for the certificate"
)
@optgroup.option(
    "-d",
    "--days-valid",
    default=365,
    help="Number of days the certificate is valid for",
)
@optgroup.group("Optional Parameters")
@optgroup.option(
    "-u",
    "--username",
    default="admin",
    help="Infoblox admin username",
)
@optgroup.option(
    "-w",
    "--wapi-ver",
    default="2.11",
    show_default=True,
    help="Infoblox WAPI version",
)
@optgroup.group("Optional Certificate Parameters")
@optgroup.option(
    "-a",
    "--algorithm",
    default="SHA-256",
    type=ALGORITHMS,
    help="The digest algorithm",
)
@optgroup.option(
    "--certificate-usage",
    default="ADMIN",
    type=USAGES,
    help="Certificate Usage",
)
@optgroup.option("-c", "--comment", help="Certificate comment")
@optgroup.option("--country", default="US", help="Certificate country")
@optgroup.option("-e", "--email", help="Certificate email address")
@optgroup.option("-k", "--key-size", default=2048, help="Certificate key size")
@optgroup.option("-l", "--locality", help="Certificate locality")
@optgroup.option("-o", "--organization", help="Certificate organization")
@optgroup.option("--ou", help="Certificate organizational unit")
@optgroup.option("-s", "--state", help="Certificate state")
@optgroup.option(
    "--san", help="Certificate subject alternative name(s) as [TYPE/VALUE,...]"
)
@optgroup.group("Logging Parameters")
@optgroup.option("--debug", is_flag=True, help="enable verbose debug output")
def selfsign(
    grid_mgr: str,
    common_name: str,
    member: str,
    days_valid: int,
    username: str,
    wapi_ver: str,
    algorithm: str,
    certificate_usage: str,
    comment: str,
    country: str,
    email: str,
    key_size: int,
    locality: str,
    organization: str,
    ou: str,
    state: str,
    san: str,
    debug: bool,
):
    if debug:
        increase_log_level()

    wapi.grid_mgr = grid_mgr
    wapi.wapi_ver = wapi_ver
    password = getpass.getpass(f"Enter password for [{username}]: ")

    try:
        wapi.connect(username=username, password=password)
    except WapiRequestException as err:
        log.error(err)
        sys.exit(1)
    else:
        log.info("connected to Infoblox grid manager %s", wapi.grid_mgr)

    subject_alt_names = san.split(",") if san else None
    if subject_alt_names:
        new_list = []
        for san in subject_alt_names:
            san_type, san_value = san.split("/")
            if san_type not in ["DNS", "IP", "URI", "EMAIL"]:
                log.error(f"Invalid subject alternative name type: {san_type}")
                sys.exit(1)
            new_list.append({"type": san_type, "value": san_value})
        subject_alt_names = new_list

    try:
        wapi.generate_selfsigned_cert(
            cn=common_name,
            member=member,
            days_valid=days_valid,
            algorithm=algorithm,
            certificate_usage=certificate_usage,
            comment=comment,
            country=country,
            email=email,
            key_size=key_size,
            locality=locality,
            org=organization,
            org_unit=ou,
            state=state,
            subject_alternative_names=subject_alt_names,
        )
    except WapiRequestException as err:
        log.error(err)
        sys.exit(1)

    log.info("finished!")
    sys.exit()


@cli.command()
@optgroup.group("Required Parameters")
@optgroup.option(
    "-g", "--grid-mgr", required=True, help="Infoblox Grid Manager"
)
@optgroup.option(
    "-n", "--common-name", required=True, help="Common Name for the certificate"
)
@optgroup.option(
    "-m", "--member", required=True, help="Member for the certificate"
)
@optgroup.group("Optional Parameters")
@optgroup.option(
    "-u",
    "--username",
    default="admin",
    help="Infoblox admin username",
)
@optgroup.option(
    "-w",
    "--wapi-ver",
    default="2.11",
    show_default=True,
    help="Infoblox WAPI version",
)
@optgroup.group("Optional Certificate Parameters")
@optgroup.option(
    "-a",
    "--algorithm",
    default="SHA-256",
    type=ALGORITHMS,
    help="The digest algorithm",
)
@optgroup.option(
    "--certificate-usage",
    default="ADMIN",
    type=USAGES,
    help="Certificate Usage",
)
@optgroup.option("-c", "--comment", help="Certificate comment")
@optgroup.option("--country", default="US", help="Certificate country")
@optgroup.option("-e", "--email", help="Certificate email address")
@optgroup.option("-k", "--key-size", default=2048, help="Certificate key size")
@optgroup.option("-l", "--locality", help="Certificate locality")
@optgroup.option("-o", "--organization", help="Certificate organization")
@optgroup.option("--ou", help="Certificate organizational unit")
@optgroup.option("-s", "--state", help="Certificate state")
@optgroup.option(
    "--san", help="Certificate subject alternative name(s) as [TYPE/VALUE,...]"
)
@optgroup.group("Logging Parameters")
@optgroup.option("--debug", is_flag=True, help="enable verbose debug output")
def gencsr(
    grid_mgr: str,
    common_name: str,
    member: str,
    username: str,
    wapi_ver: str,
    algorithm: str,
    certificate_usage: str,
    comment: str,
    country: str,
    email: str,
    key_size: int,
    locality: str,
    organization: str,
    ou: str,
    state: str,
    san: str,
    debug: bool,
) -> None:
    if debug:
        increase_log_level()

    wapi.grid_mgr = grid_mgr
    wapi.wapi_ver = wapi_ver
    password = getpass.getpass(f"Enter password for [{username}]: ")

    try:
        wapi.connect(username=username, password=password)
    except WapiRequestException as err:
        log.error(err)
        sys.exit(1)
    else:
        log.info("connected to Infoblox grid manager %s", wapi.grid_mgr)

    subject_alt_names = san.split(",") if san else None
    if subject_alt_names:
        new_list = []
        for san in subject_alt_names:
            san_type, san_value = san.split("/")
            if san_type not in ["DNS", "IP", "URI", "EMAIL"]:
                log.error(f"Invalid subject alternative name type: {san_type}")
                sys.exit(1)
            new_list.append({"type": san_type, "value": san_value})
        subject_alt_names = new_list

    try:
        wapi.generate_csr(
            cn=common_name,
            member=member,
            algorithm=algorithm,
            certificate_usage=certificate_usage,
            comment=comment,
            country=country,
            email=email,
            key_size=key_size,
            locality=locality,
            org=organization,
            org_unit=ou,
            state=state,
            subject_alternative_names=subject_alt_names,
        )
    except WapiRequestException as err:
        log.error(err)
        sys.exit(1)

    log.info("finished!")
    sys.exit()


if __name__ == "__main__":
    asyncio.run(cli())



from typing import List

from pydantic import (
    BaseModel,
    Field,
    ConfigDict,
    field_serializer,
    IPvAnyAddress,
)

from .enums import ZoneFormatTypeEnum, ImportActionEnum


def get_zone_format(zone_name: str) -> ZoneFormatTypeEnum:
    if "in-addr.arpa" in zone_name.lower():
        zone_format = ZoneFormatTypeEnum.IPV4
    elif "ip6.arpa" in zone_name.lower():
        zone_format = ZoneFormatTypeEnum.IPV6
    else:
        zone_format = ZoneFormatTypeEnum.FORWARD
    return zone_format


class MemberDns(BaseModel):
    model_config = ConfigDict(extra="allow")

    memberdns: str = Field(
        "memberdns",
        frozen=True,
        serialization_alias="header-memberdns",
        description="CSV header for memberdns",
    )
    import_action: ImportActionEnum | None = Field(
        serialization_alias="import-action",
        default=None,
        description="CSV custom import action",
    )
    parent: str = Field(..., description="Parent FQDN Member")
    dns_over_mgmt: bool | None = Field(
        None, description="DNS over management interface"
    )
    dns_over_lan2: bool | None = Field(
        None, description="DNS over LAN2 interface"
    )
    minimal_response: bool | None = Field(
        None, description="Minimal response flag"
    )
    forwarders_only: bool | None = Field(
        None, description="Forwarders only flag"
    )
    allow_forwarder: List[str | None] = Field(
        None, description="List of forwarders"
    )
    member_view_nats: List[str | None] = Field(
        None, description="List of member view NATs VIEW/INTERFACE/IP"
    )
    enable_notify_source_port: bool | None = Field(
        None, description="Enable notify source port"
    )
    notify_source_port: int | None = Field(
        None, description="Notify source port"
    )
    enable_query_source_port: bool | None = Field(
        None, description="Enable query source port"
    )
    query_source_port: int | None = Field(
        None, description="Query source port"
    )
    lame_ttl: int | None = Field(
        600, gt=0, lt=1800, description="Lame TTL"
    )
    auto_sort_views: bool | None = Field(
        None, description="Auto sort views flag"
    )
    member_views: List[str | None] = Field(None, description="Member views")
    allow_transfer: List[str | None] = Field(
        None,
        description="List of allowed transfer servers ITEM/Allow or ITEM/Deny",
    )
    excluded_servers: List[IPvAnyAddress | None] = Field(
        None, description="List of excluded transfer servers"
    )
    zone_transfer_format_option: str | None = Field(
        None, description="Zone transfer format option"
    )
    recursion_enabled: bool | None = Field(
        None, description="Recursion enabled flag"
    )
    allow_query: List[str | None] = Field(
        None, description="List of allowed query servers"
    )
    allow_recursive_query: List[str | None] = Field(
        None, description="List of allowed recursive query servers"
    )
    limit_concurrent_recursive_clients: bool | None = Field(
        None, description="Limit concurrent recursive clients flag"
    )
    concurrent_recursive_clients: int | None = Field(
        1000, description="Concurrent recursive clients"
    )
    allow_update: List[str | None] = Field(
        None, description="List of allowed update servers"
    )
    allow_gss_tsig_zone_updates: bool | None = Field(
        None, description="Allow GSS TSIG zone updates flag"
    )
    allow_update_forwarding: bool | None = Field(
        None, description="Allow update forwarding flag"
    )
    enable_custom_root_server: bool | None = Field(
        None, description="Enable custom root server"
    )
    root_name_servers: List[str | None] = Field(
        None, description="List of custom root name servers"
    )
    enable_blackhole: bool | None = Field(
        None, description="Enable blackhole flag"
    )
    blackhole: List[str | None] = Field(
        None, description="List of blackhole servers"
    )
    notify_delay: int | None = Field(
        None, ge=5, le=86400, description="Notify delay in seconds"
    )
    enable_nxdomain_redirect: bool | None = Field(
        None, description="Enable nxdomain redirect flag"
    )
    nxdomain_redirect_addresses: List[str | None] = Field(
        None, description="List of nxdomain redirect addresses"
    )
    nxdomain_redirect_ttl: int | None = Field(
        None, description="TTL for nxdomain redirect"
    )
    nxdomain_log_query: bool | None = Field(
        None, description="Enable nxdomain log query flag"
    )
    nxdomain_rulesets: List[str | None] = Field(
        None, description="List of nxdomain rulesets"
    )
    enable_blacklist: bool | None = Field(
        None, description="Enable blacklist flag"
    )
    blacklist_redirect_addresses: List[str | None] = Field(
        None, description="List of blacklist redirect addresses"
    )
    blacklist_action: str | None = Field(
        None, description="Blacklist action i.e. Refuse"
    )
    blacklist_redirect_ttl: int | None = Field(
        None, description="TTL for blacklist redirect"
    )
    blacklist_log_query: bool | None = Field(
        None, description="Enable blacklist log query flag"
    )
    blacklist_rulesets: List[str | None] = Field(
        None, description="List of blacklist rulesets"
    )
    enable_dns64: bool | None = Field(None, description="Enable DNS64 flag")
    dns64_groups: List[str | None] = Field(
        None, description="List of DNS64 groups"
    )
    max_cached_lifetime: int | None = Field(
        86400, ge=60, le=86400, description="Max cached lifetime in seconds"
    )
    dns_over_v6_mgmt: bool | None = Field(
        None, description="DNS over v6 management interface"
    )
    dns_over_v6_lan2: bool | None = Field(
        None, description="DNS over v6 LAN2 interface"
    )
    filter_aaaa: bool | None = Field(
        None, description="Enable filter AAAA flag"
    )
    filter_aaaa_list: List[str | None] = Field(
        None,
        description="List of filter AAAA addresses '12.0.0.12/Deny,10.0.0.0/8/Allow,NACL/Allow'",
    )
    dns_over_v6_lan: bool | None = Field(
        None, description="DNS over v6 LAN interface"
    )
    copy_xfer_to_notify: bool | None = Field(
        None, description="Enable copy transfer to notify flag"
    )
    transfers_in: int | None = Field(
        10, ge=10, le=100, description="Max transfers in"
    )
    transfers_out: int | None = Field(
        10, ge=10, le=100, description="Max transfers out"
    )
    transfers_per_ns: int | None = Field(
        None, ge=2, le=100, description="Max transfers per NS"
    )
    serial_query_rate: int | None = Field(
        20, ge=20, le=100, description="Serial query rate"
    )
    max_cache_ttl: int | None = Field(
        86400, description="Max cache TTL in seconds"
    )
    max_ncache_ttl: int | None = Field(
        10800, le=604800, description="Max ncache TTL in seconds"
    )
    disable_edns: bool | None = Field(None, description="Disable EDNS flag")
    query_rewrite_enabled: bool | None = Field(
        None, description="Enable query rewrite flag"
    )
    rpz_drop_ip_rule_enabled: bool | None = Field(
        None, description="Enable RPZ drop IP rule flag"
    )
    rpz_drop_ip_rule_min_prefix_length_ipv4: int | None = Field(
        29, ge=0, le=32, description="RPZ drop IP rule min prefix length IPv4"
    )
    rpz_drop_ip_rule_min_prefix_length_ipv6: int | None = Field(
        112, ge=0, le=128, description="RPZ drop IP rule min prefix length IPv6"
    )
    atc_forwarding_enable: bool | None = Field(
        None,
        description="Enable recursive queries forwarding to Advanced DNS Threat Defense Cloud flag",
    )
    atc_forwarding_access_key: str | None = Field(
        None, description="Advanced DNS Threat Defense Cloud access key"
    )
    atc_forwarding_resolver_address: IPvAnyAddress | None = Field(
        None, description="Advanced DNS Threat Defense Cloud resolver address"
    )
    atc_forwarding_forward_first: bool | None = Field(
        None,
        description="Fallback to Advanced DNS Threat Defense Cloud forward first flag",
    )

    @staticmethod
    def list_to_csv(items: List[str | None]) -> str | None:
        if not items:
            return None
        return ",".join(items)

    @field_serializer(
        "allow_forwarder",
        "member_view_nats",
        "member_views",
        "allow_transfer",
        "excluded_servers",
        "allow_query",
        "allow_recursive_query",
        "allow_update",
        "root_name_servers",
        "blackhole",
        "nxdomain_redirect_addresses",
        "nxdomain_rulesets",
        "blacklist_redirect_addresses",
        "blacklist_rulesets",
        "dns64_groups",
        "filter_aaaa_list",
        when_used="always",
    )
    def serialize_list_fields(
        self, values: List[str | None]
    ) -> str | None:
        return self.list_to_csv(values)


class DnsView(BaseModel):
    model_config = ConfigDict(extra="allow")

    view: str = Field(
        "view",
        frozen=True,
        serialization_alias="header-view",
        description="CSV header for view object",
    )
    import_action: ImportActionEnum | None = Field(
        None,
        serialization_alias="import-action",
        description="CSV Custom import action",
    )
    name: str = Field(..., description="View name")
    new_name: str | None = Field(
        None, serialization_alias="_new_name", description="New view name"
    )
    comment: str | None = Field(
        None, description="Optional view comment field"
    )
    network_view: str | None = Field(None, description="Network view name")
    disable: bool | None = Field(None, description="Disable view flag")
    recursion: bool | None = Field(None, description="Enable recursion flag")
    root_name_server_type: str | None = Field(
        None, description="Root name server type, i.e. Custom"
    )
    match_clients: List[str | None] = Field(
        None, description="List of clients to match [IP|NET|ACL,...]"
    )
    match_destinations: List[str | None] = Field(
        None, description="List of destinations to match [IP|NET|ACL,...]"
    )
    custom_root_name_servers: List[str | None] = Field(
        None, description="List of custom root name servers [FQDN/IP,...]"
    )
    lame_ttl: int | None = Field(None, description="Lame TTL value")
    nxdomain_redirect: bool | None = Field(
        None, description="Enable NXDOMAIN redirect flag"
    )
    nxdomain_redirect_addresses: List[str | None] = Field(
        None, description="List of NXDOMAIN redirect addresses [IP,...]"
    )
    nxdomain_redirect_ttl: int | None = Field(
        None, description="NXDOMAIN redirect TTL value"
    )
    nxdomain_log_query: bool | None = Field(
        None, description="Enable NXDOMAIN log query flag"
    )
    nxdomain_rulesets: List[str | None] = Field(
        None, description="List of NXDOMAIN rulesets [FQDNs,...]"
    )
    enable_blacklist: bool | None = Field(
        None, description="Enable blacklist flag"
    )
    blacklist_redirect_addresses: List[str | None] = Field(
        None, description="List of blacklist redirect addresses [IP,...]"
    )
    blacklist_action: str | None = Field(
        None, description="Blacklist action, i.e. Redirect"
    )
    blacklist_redirect_ttl: int | None = Field(
        None, description="Blacklist redirect TTL value"
    )
    blacklist_log_query: bool | None = Field(
        None, description="Enable blacklist log query flag"
    )
    blacklist_rulesets: List[str | None] = Field(
        None, description="List of blacklist rulesets [FQDNs,...]"
    )
    enable_dns64: bool | None = Field(None, description="Enable DNS64 flag")
    dns64_groups: List[str | None] = Field(
        None, description="List of DNS64 groups [GROUP,...]"
    )
    forwarders_only: bool | None = Field(
        None, description="Forwarders only flag"
    )
    forwarders: List[str | None] = Field(
        None, description="List of forwarders [IP,...]"
    )
    filter_aaaa: str | None = Field(
        None, description="Enable filter AAAA flag, i.e. Yes or No"
    )
    filter_aaaa_list: List[str | None] = Field(
        None, description="List of filter AAAA addresses [IP/Deny|Allow,...]"
    )
    max_cache_ttl: int | None = Field(
        None, description="Max cache TTL in seconds"
    )
    max_ncache_ttl: int | None = Field(
        None, description="Max ncache TTL in seconds"
    )
    rpz_drop_ip_rule_enabled: bool | None = Field(
        None, description="Enable RPZ drop IP rule flag"
    )
    rpz_drop_ip_rule_min_prefix_length_ipv4: int | None = Field(
        None, description="RPZ drop IP rule min prefix length IPv4"
    )
    rpz_drop_ip_rule_min_prefix_length_ipv6: int | None = Field(
        None, description="RPZ drop IP rule min prefix length IPv6"
    )

    def add_property(self, prop: str, value: str):
        if prop.startswith("EA-") or prop.startswith("ADMGRP-"):
            self.__setattr__(prop, value)
        else:
            raise Exception(f"Invalid field name: {prop}")


class AuthZone(BaseModel):
    """
    Represents an AuthZone with configurations and settings for managing DNS zones.

    This class models an AuthZone, providing attributes for various DNS zone
    settings like FQDN, zone format, views, primary and secondary servers,
    SOA configuration, query, update, transfer permissions, and more. It also
    allows for handling custom properties and field serialization. Suitable
    for managing detailed DNS zone data and related configurations.

    Attributes:
        model_config (ConfigDict): Model configuration allowing extra fields.
        authzone (str): CSV header for an authzone object (immutable).
        import_action (ImportActionEnum | None): CSV Custom import action.
        fqdn (str): Fully Qualified Domain Name of the zone.
        zone_format (ZoneFormatTypeEnum): Zone format enumeration.
        view (str | None): View name associated with the zone.
        prefix (str | None): RFC2317 classless reverse zone Prefix.
        new_prefix (str | None): New RFC2317 classless reverse zone Prefix.
        is_multimaster (bool | None): Flag indicating multimaster setup.
        grid_primaries (List[str | None]): List of grid primary servers.
        external_primaries (List[str | None]): List of external primary servers.
        grid_secondaries (List[str | None]): List of grid secondary servers.
        external_secondaries (List[str | None]): List of external secondary servers.
        ns_group (str | None): Name of the NS group.
        comment (str | None): Optional comment on the DNS zone.
        disabled (bool | None): Flag indicating whether the zone is disabled.
        create_underscore_zones (bool | None): Flag to create underscore zones.
        allow_active_dir (List[str | None]): List of allowed Active Directory servers.
        soa_refresh (int | None): SOA refresh value in seconds.
        soa_retry (int | None): SOA retry value in seconds.
        soa_expire (int | None): SOA expire value in seconds.
        soa_default_ttl (int | None): Default TTL value for SOA.
        soa_negative_ttl (int | None): Negative TTL value for SOA.
        soa_mnames (List[str | None]): List of SOA primary names.
        soa_email (str | None): Email address for the SOA records.
        soa_serial_number (int | None): Serial number of the SOA.
        disable_forwarding (bool | None): Flag to disable forwarding.
        allow_update_forwarding (bool | None): Flag to permit update forwarding.
        update_forwarding (List[str | None]): Update a forwarding server list.
        allow_transfer (List[str | None]): List of allowed transfer servers.
        allow_update (List[str | None]): List of allowed update servers.
        allow_query (List[str | None]): List of allowed query servers.
        notify_delay (int | None): Notify delay in seconds (5-86400).

    Methods:
        list_to_csv(items: List[str | None]) -> str | None:
            Converts a list of strings to a CSV-formatted string.
            Returns None if the list is empty or None.

        serialize_list_fields(values: List[str | None]) -> str | None:
            Serializes specific list fields to a CSV format using list_to_csv.
            Applicable on attributes such as `grid_primaries`, `external_primaries`,
            `grid_secondaries`, and others explicitly defined in the decorator.

        add_property(prop: str, value: str):
            Adds a custom property to the instance if the property name
            starts with "EA-" or "ADMGRP-". Raises an exception for invalid names.
    """

    model_config = ConfigDict(extra="allow")

    authzone: str = Field(
        "authzone",
        frozen=True,
        serialization_alias="header-authzone",
        description="CSV header for authzone object",
    )
    import_action: ImportActionEnum | None = Field(
        None,
        serialization_alias="import-action",
        description="CSV Custom import action",
    )
    fqdn: str = Field(..., description="FQDN zone name")
    zone_format: ZoneFormatTypeEnum = Field(..., description="Zone format")
    view: str | None = Field(None, description="View name")
    prefix: str | None = Field(
        None, description="RFC2317 classless reverse zone Prefix"
    )
    new_prefix: str | None = Field(
        None,
        serialization_alias="_new_prefix",
        description="New RFC2317 classless reverse zone Prefix",
    )
    is_multimaster: bool | None = Field(
        None, description="Is multimaster flag"
    )
    grid_primaries: List[str | None] = Field(
        None, description="Grid primary servers [MEMBER/IS_STEALTH,...]"
    )
    external_primaries: List[str | None] = Field(
        None,
        description="External primary servers [FQDN/IP/USE_2X_TSIG/USE_TSIG/NAME/KEY/ALGO,...]",
    )
    grid_secondaries: List[str | None] = Field(
        None,
        description="Grid secondary servers [FQDN/STEALTH/LEAD/GRID_SYNC,...] ",
    )
    external_secondaries: List[str | None] = Field(
        None,
        description="External secondary servers [FQDN/IP/USE_2X_TSIG/USE_TSIG/NAME/KEY/ALGO,...]",
    )
    ns_group: str | None = Field(None, description="NS group name")
    comment: str | None = Field(None, description="Optional comment")
    disabled: bool | None = Field(None, description="Disabled flag")
    create_underscore_zones: bool | None = Field(
        None, description="Create underscore zones flag"
    )
    allow_active_dir: List[str | None] = Field(
        None, description="List of allowed Active Directory servers"
    )
    soa_refresh: int | None = Field(
        None, description="SOA refresh value"
    )
    soa_retry: int | None = Field(
        None, description="SOA retry value"
    )
    soa_expire: int | None = Field(
        None, description="SOA expire value"
    )
    soa_default_ttl: int | None = Field(
        None, description="SOA default TTL value"
    )
    soa_negative_ttl: int | None = Field(
        None, description="SOA negative TTL value"
    )
    soa_mnames: List[str | None] = Field(
        None, description="List of SOA mnames [ZONE/FQDN,...]"
    )
    soa_email: str | None = Field(None, description="SOA email address")
    soa_serial_number: int | None = Field(
        None, description="SOA serial number"
    )
    disable_forwarding: bool | None = Field(
        None, description="Disable forwarding or empty forwarders flag"
    )
    allow_update_forwarding: bool | None = Field(
        None, description="Allow update forwarding flag"
    )
    update_forwarding: List[str | None] = Field(
        None,
        description="List of update forwarding servers [ITEM/PERMISSION,...]",
    )
    allow_transfer: List[str | None] = Field(
        None,
        description="List of allowed transfer servers [ITEM/PERMISSION,...]",
    )
    allow_update: List[str | None] = Field(
        None, description="List of allowed update servers [ITEM/PERMISSION,...]"
    )
    allow_query: List[str | None] = Field(
        None, description="List of allowed query servers [ITEM/PERMISSION,...]"
    )
    notify_delay: int | None = Field(
        None, ge=5, le=86400, description="Notify delay in seconds"
    )

    @staticmethod
    def list_to_csv(items: List[str | None]) -> str | None:
        if not items:
            return None
        return ",".join(items)

    @field_serializer(
        "grid_primaries",
        "external_primaries",
        "grid_secondaries",
        "external_secondaries",
        "allow_active_dir",
        "soa_mnames",
        "update_forwarding",
        "allow_transfer",
        "allow_update",
        "allow_query",
    )
    def serialize_list_fields(
        self, values: List[str | None]
    ) -> str | None:
        return self.list_to_csv(values)

    def add_property(self, prop: str, value: str):
        if prop.startswith("EA-") or prop.startswith("ADMGRP-"):
            self.__setattr__(prop, value)
        else:
            raise Exception(f"Invalid field name: {prop}")


class ForwardZone(BaseModel):
    """
    Represents a ForwardZone that defines the configuration and properties
    related to DNS forward zones.

    This class models the configurations and data for a DNS forward zone. It
    stores various metadata attributes such as forwarders, disabled flags,
    comments, zone format, and more. It provides mechanisms for serialization
    and utility methods to handle specific data transformations. It also
    includes validation checks for certain property names during data addition.

    Attributes:
        forwardzone (str): Header for the forwardzone object.
        import_action (ImportActionEnum | None): Custom import action for CSV files.
        fqdn (str): Fully Qualified Domain Name (FQDN) of the zone.
        view (str | None): View name associated with the zone.
        zone_format (ZoneFormatTypeEnum): Defines the format of the zone.
        prefix (str | None): RFC2317 classless reverse zone prefix.
        disabled (bool | None): Indicates whether the zone is disabled.
        comment (str | None): Optional comment or description for the zone.
        forward_to (List[str | None]): List of forwarders in FQDN/IP format.
        forwarding_servers (List[str | None]): List of forwarding servers in FQDN format.
        forwarders_only (bool | None): Indicates if only forwarders are used.
        ns_group (str | None): Group of nameservers for forwarding members.
        ns_group_external (str | None): External group name for forward-to nameservers.
        disable_ns_generation (bool | None): Indicates if the NS generation is disabled.
    """

    model_config = ConfigDict(extra="allow")

    forwardzone: str = Field(
        "forwardzone",
        frozen=True,
        serialization_alias="header-forwardzone",
        description="Header for forwardzone object",
    )
    import_action: ImportActionEnum | None = Field(
        None,
        serialization_alias="import-action",
        description="CSV Custom import action",
    )
    fqdn: str = Field(..., min_length=1, description="FQDN zone name")
    view: str | None = Field(None, description="View name")
    zone_format: ZoneFormatTypeEnum = Field(..., description="Zone format")
    prefix: str | None = Field(
        None, description="RFC2317 classless reverse zone Prefix"
    )
    disabled: bool | None = Field(None, description="Disabled flag")
    comment: str | None = Field(None, description="Optional comment")
    forward_to: List[str | None] = Field(
        None, description="List of forwarders [FQDN/IP,...]"
    )
    forwarding_servers: List[str | None] = Field(
        None, description="List of forwarding servers [FQDN,...]"
    )
    forwarders_only: bool | None = Field(
        True, description="Forwarders only flag"
    )
    ns_group: str | None = Field(
        None, description="Forwarding Members NS Group"
    )
    ns_group_external: str | None = Field(
        None, description="Forward-to NS group name"
    )
    disable_ns_generation: bool | None = Field(
        True, description="Disable NS generation flag"
    )

    @staticmethod
    def list_to_csv(items: List[str | None]) -> str | None:
        if not items:
            return None
        return ",".join(items)

    @field_serializer("forward_to", "forwarding_servers", when_used="always")
    def serialize_list_fields(
        self, values: List[str | None]
    ) -> str | None:
        return self.list_to_csv(values)

    def add_property(self, prop: str, value: str):
        if prop.startswith("EA-") or prop.startswith("ADMGRP-"):
            self.__setattr__(prop, value)
        else:
            raise Exception(f"Invalid field name: {prop}")


class StubZone(BaseModel):
    """
    StubZone class used for representing and managing stub zone configurations. It extends
    the BaseModel class and uses Pydantic for data validation and serialization. The class
    primarily focuses on configurations related to DNS stub zones, such as FQDN, zone format,
    view, forwarding rules, and related settings.

    Attributes:
        stubzone (str): Header for a stubzone object. Frozen and aliased for serialization.
        import_action (ImportActionEnum | None): CSV custom import action. Aliased
            for serialization.
        fqdn (str): Fully Qualified Domain Name (FQDN) of the zone. Must meet a
            minimum length requirement of 1.
        view (str | None): Name of the DNS view.
        zone_format (ZoneFormatTypeEnum): Format of the DNS zone.
        prefix (int | None): RFC2317 classless reverse zone prefix. Must
            have a value between 24 and 32.
        disabled (bool | None): Indicates whether the stub zone is disabled.
        comment (str | None): Optional comment for additional information.
        disable_forwarding (bool | None): Indicates if forwarding is disabled.
        stub_from (List[str | None]): List of stub-from servers represented as
            FQDN or IP addresses.
        stub_members (List[str | None]): List of stub-members servers represented
            as FQDN values.
        ns_group (str | None): Specifies the Stub Members NS Group.
        ns_group_external (str | None): Specifies the NS group name for stub-from
            servers.

    Methods:
        list_to_csv(items: List[str | None]) -> str | None
            Converts a list of strings to a single CSV string. Returns None if the
            input is empty or None.

        serialize_list_fields(values: List[str | None]) -> str | None
            Field serializer for stub_from and stub_members attributes. Serializes
            lists into CSV strings for consistent representation during serialization.

        add_property(prop: str, value: str)
            Dynamically adds a property to the instance if its name starts with
            a valid prefix ("EA-" or "ADMGRP-"). Raises an Exception for invalid
            property names.
    """

    model_config = ConfigDict(extra="allow")

    stubzone: str = Field(
        "stubzone",
        frozen=True,
        serialization_alias="header-stubzone",
        description="Header for stubzone object",
    )
    import_action: ImportActionEnum | None = Field(
        None,
        serialization_alias="import-action",
        description="CSV Custom import action",
    )
    fqdn: str = Field(..., min_length=1, description="FQDN zone name")
    view: str | None = Field(None, description="View name")
    zone_format: ZoneFormatTypeEnum = Field(..., description="Zone format")
    prefix: int | None = Field(
        None, description="RFC2317 classless reverse zone Prefix", ge=24, le=32
    )
    disabled: bool | None = Field(None, description="Disabled flag")
    comment: str | None = Field(None, description="Optional comment")
    disable_forwarding: bool | None = Field(
        None, description="Disable forwarding flag"
    )
    stub_from: List[str | None] = Field(
        None, description="List of stub-from servers [FQDN/IP,...]"
    )
    stub_members: List[str | None] = Field(
        None, description="List of stub-members servers [FQDN,...]"
    )
    ns_group: str | None = Field(None, description="Stub Members NS Group")
    ns_group_external: str | None = Field(
        None, description="Stub-from NS group name"
    )

    @staticmethod
    def list_to_csv(items: List[str | None]) -> str | None:
        if not items:
            return None
        return ",".join(items)

    @field_serializer("stub_from", "stub_members", when_used="always")
    def serialize_list_fields(
        self, values: List[str | None]
    ) -> str | None:
        return self.list_to_csv(values)

    def add_property(self, prop: str, value: str):
        if prop.startswith("EA-") or prop.startswith("ADMGRP-"):
            self.__setattr__(prop, value)
        else:
            raise Exception(f"Invalid field name: {prop}")


class NsGroup(BaseModel):
    model_config = ConfigDict(extra="allow")

    nsgroup: str = Field(
        "nsgroup",
        frozen=True,
        serialization_alias="header-nsgroup",
        description="CSV header for nsgroup object",
    )
    import_action: ImportActionEnum | None = Field(
        None,
        serialization_alias="import-action",
        description="CSV Custom import action",
    )
    group_name: str = Field(..., description="NS group name")
    new_group_name: str | None = Field(
        None,
        serialization_alias="_new_group_name",
        description="New NS group name",
    )
    grid_primaries: List[str | None] = Field(
        None, description="Grid primary servers [MEMBER/IS_STEALTH,...]"
    )
    external_primaries: List[str | None] = Field(
        None,
        description="External primary servers [FQDN/IP/STEALTH/USE_2X_TSIG/USE_TSIG/NAME/KEY,...]",
    )
    external_secondaries: List[str | None] = Field(
        None,
        description="External secondary servers [FQDN/IP/STEALTH/USE_2X_TSIG/USE_TSIG/NAME/KEY,...]",
    )
    grid_secondaries: List[str | None] = Field(
        None,
        description="Grid secondary servers [FQDN/IS_STEALTH/LEAD/GRID_SYNC,...] ",
    )
    is_grid_default: bool | None = Field(
        None, description="Is grid default flag"
    )
    comment: str | None = Field(None, description="Optional comment")

    @staticmethod
    def list_to_csv(items: List[str | None]) -> str | None:
        if not items:
            return None
        return ",".join(items)

    @field_serializer(
        "grid_primaries",
        "external_primaries",
        "external_secondaries",
        "grid_secondaries",
        when_used="always",
    )
    def serialize_list_fields(
        self, values: List[str | None]
    ) -> str | None:
        return self.list_to_csv(values)

    def add_property(self, prop: str, value: str):
        if prop.startswith("EA-"):
            self.__setattr__(prop, value)
        else:
            raise Exception(f"Invalid field name: {prop}")


class DelegatedZone(BaseModel):
    """
    Represents a Delegated Zone model with configuration and related properties.

    The DelegatedZone class is designed to represent the structure and behaviors
    associated with configuration data for delegated zones in a DNS system. It uses
    strict configuration enforcement via field definitions and type hints, allowing
    extra fields through a flexible model configuration. The class includes features
    such as property-based serialization and validation for dynamically added attributes.

    Attributes:
        model_config: Configuration that allows additional fields beyond those explicitly
            defined in the model.
        delegatedzone: CSV Header for a delegatedzone object.
        import_action: Custom import action for the zone, optional.
        fqdn: Fully Qualified Domain Name (FQDN) for the zone.
        view: Name of the view associated with the zone, optional.
        zone_format: Format of the zone (e.g., forward, reverse).
        prefix: Prefix value for RFC2317 classless reverse zones, optional.
        disabled: Flag indicating whether the zone is disabled, optional.
        comment: Optional comment for the zone, optional.
        delegate_to: A list of delegated servers expressed in FQDN or IP format, optional.
        delegated_ttl: Time To Live (TTL) value for the delegation, optional.
        ns_group: Name of the NS Group for delegated members, optional.
        new_prefix: New prefix for RFC2317 classless reverse zones, optional.
        ddns_protected: Indicates whether the record is protected by DDNS, optional.
        ddns_principal: The principal used for DDNS operations, optional.

    Methods:
        serialize_delegate_to: Serializes the list of delegated servers into a string format
            separated by commas if the list exists; otherwise, returns None.
        add_property: Adds a dynamic property to the instance if it follows defined naming
            conventions (e.g., starts with 'EA-' or 'ADMGRP-'). Raises an exception if the
            naming convention is not followed.
    """

    model_config = ConfigDict(extra="allow")

    delegatedzone: str = Field(
        "delegatedzone",
        frozen=True,
        serialization_alias="header-delegatedzone",
        description="CSV Header for delegatedzone object",
    )
    import_action: ImportActionEnum | None = Field(
        None,
        serialization_alias="import-action",
        description="CSV Custom import action",
    )
    fqdn: str = Field(..., description="FQDN zone name")
    view: str | None = Field(None, description="View name")
    zone_format: ZoneFormatTypeEnum = Field(..., description="Zone format")
    prefix: int | None = Field(
        None, description="RFC2317 classless reverse zone Prefix"
    )
    disabled: bool | None = Field(None, description="Disabled flag")
    comment: str | None = Field(None, description="Optional comment")
    delegate_to: List[str | None] = Field(
        None, description="List of delegated servers [FQDN/IP,...]"
    )
    delegated_ttl: int | None = Field(
        None, description="Delegated TTL"
    )
    ns_group: str | None = Field(
        None, description="Delegated Members NS Group"
    )
    new_prefix: int | None = Field(
        None,
        serialization_alias="_new_prefix",
        description="New RFC2317 classless reverse zone Prefix",
    )
    ddns_protected: bool | None = Field(
        None, description="DDNS protected flag"
    )
    ddns_principal: str | None = Field(None, description="DDNS principal")

    @field_serializer("delegate_to", when_used="always")
    def serialize_delegate_to(self, items: List[str | None]) -> str | None:
        if not items:
            return None
        return ",".join(items)

    def add_property(self, prop: str, value: str):
        if prop.startswith("EA-") or prop.startswith("ADMGRP-"):
            self.__setattr__(prop, value)
        else:
            raise Exception(f"Invalid field name: {prop}")


class DelegationNsGroup(BaseModel):
    model_config = ConfigDict(extra="allow")

    delegationnsgroup: str = Field(
        "delegationnsgroup",
        frozen=True,
        serialization_alias="header-delegationnsgroup",
        description="Header for delegationnsgroup object",
    )
    import_action: ImportActionEnum | None = Field(
        None,
        serialization_alias="import-action",
        description="CSV Custom import action",
    )
    group_name: str = Field(..., description="NS group name")
    new_group_name: str | None = Field(
        None,
        serialization_alias="_new_group_name",
        description="New NS group name",
    )
    delegate_to: List[str | None] = Field(
        None, description="List of delegated servers [FQDN/IP,...]"
    )
    comment: str | None = Field(None, description="Optional comment")

    @field_serializer("delegate_to", when_used="always")
    def serialize_delegate_to(self, items: List[str | None]) -> str | None:
        if not items:
            return None
        return ",".join(items)


class ForwardingMemberNsGroup(BaseModel):
    model_config = ConfigDict(extra="allow")

    forwardingmembernsgroup: str = Field(
        "forwardingmembernsgroup",
        frozen=True,
        serialization_alias="header-forwardingmembernsgroup",
        description="Header for forwardingmembernsgroup object",
    )
    import_action: ImportActionEnum | None = Field(
        None,
        serialization_alias="import-action",
        description="CSV Custom import action",
    )
    group_name: str = Field(..., description="NS group name")
    new_group_name: str | None = Field(
        None,
        serialization_alias="_new_group_name",
        description="New NS group name",
    )
    comment: str | None = Field(None, description="Optional comment")
    forwarding_servers: List[str | None] = Field(
        None, description="List of forwarding servers [FQDN/IP,...]"
    )

    @field_serializer("forwarding_servers", when_used="always")
    def serialize_forwarding_servers(
        self, items: List[str | None]
    ) -> str | None:
        if not items:
            return None
        return ",".join(items)

    def add_property(self, prop: str, value: str):
        if prop.startswith("EA-"):
            self.__setattr__(prop, value)
        else:
            raise Exception(f"Invalid field name: {prop}")


class ForwardStubServerNsGroup(BaseModel):
    model_config = ConfigDict(extra="allow")

    forwardstubservernsgroup: str = Field(
        "forwardstubservernsgroup",
        frozen=True,
        serialization_alias="header-forwardstubservernsgroup",
        description="Header for forwardstubservernsgroup object",
    )
    import_action: ImportActionEnum | None = Field(
        None,
        serialization_alias="import-action",
        description="CSV Custom import action",
    )
    group_name: str = Field(..., description="NS group name")
    new_group_name: str | None = Field(
        None,
        serialization_alias="_new_group_name",
        description="New NS group name",
    )
    comment: str | None = Field(None, description="Optional comment")
    external_servers: List[str | None] = Field(
        None, description="List of external servers [FQDN/IP,...]"
    )

    @field_serializer("external_servers", when_used="always")
    def serialize_external_servers(
        self, items: List[str | None]
    ) -> str | None:
        if not items:
            return None
        return ",".join(items)

    def add_property(self, prop: str, value: str):
        if prop.startswith("EA-"):
            self.__setattr__(prop, value)
        else:
            raise Exception(f"Invalid field name: {prop}")


class StubMemberNsGroup(BaseModel):
    model_config = ConfigDict(extra="allow")

    stubmembernsgroup: str = Field(
        "stubmembernsgroup",
        frozen=True,
        serialization_alias="header-stubmembernsgroup",
        description="CSV Header for stubmembernsgroup objects",
    )
    import_action: ImportActionEnum | None = Field(
        None,
        serialization_alias="import-action",
        description="CSV Custom import action",
    )
    group_name: str = Field(..., description="NS group name", min_length=1)
    new_group_name: str | None = Field(
        None,
        serialization_alias="_new_group_name",
        description="New NS group name",
    )
    comment: str | None = Field(None, description="Optional comment")
    stub_members: List[str | None] = Field(
        None, description="List of stub-members servers [FQDN,...]"
    )

    @field_serializer("stub_members", when_used="always")
    def serialize_stub_members(self, items: List[str | None]) -> str | None:
        if not items:
            return None
        return ",".join(items)

    def add_property(self, prop: str, value: str):
        if prop.startswith("EA-"):
            self.__setattr__(prop, value)
        else:
            raise Exception(f"Invalid field name: {prop}")



"""
Copyright 2023 Infoblox

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import csv
import io
import logging
import os
import pprint
import subprocess
from urllib.parse import urlparse


def named_checkconf(chroot_path: str, conf_path: str) -> None:
    """
    perform named-checkconf and re-write a canonicalized copy of the named.conf file

    Args:
        chroot_path (str): Description of the first argument `chroot_path`.
        conf_path (str): Description of the second argument `conf_path`.

    Returns:
        None

    Example:
        Here you can provide an example of how to use your function.
        >>> named_checkconf("/path/to/chroot", "/path/to/conf")
    """

    logging.info("parsing %s conf file", conf_path)

    if not os.path.exists(f"{chroot_path}/{conf_path}"):
        logging.error("%s does not exist", f"{chroot_path}/{conf_path}")
        raise FileNotFoundError
    command = (
        f"sudo named-checkconf -p -t {chroot_path} {conf_path} > name.conf"
    )
    logging.debug(command)
    res = subprocess.run(
        ["sudo", "named-checkconf", "-p", "-t", chroot_path, conf_path],
        capture_output=True,
        text=True,
        check=True,
    )
    if res.stdout:
        with open("named.conf", "w", encoding="utf8") as file:
            file.write(res.stdout)
    if res.stderr:
        logging.error(res.stderr)


def named_compilezone(
    zone_name: str, zone_file: str, output_file: str, input_format: str = "text"
) -> list:
    """
    The function named_compilezone performs the named-compilezone command to canonicalize and
    rewrite a specified DNS zone file, checking for errors in the file during the process.

    Args:
        zone_name (str): The name of the DNS zone being processed.
        zone_file (str): The path to the file containing the DNS zone information to be
                         read/processed.
        output_file (str): The path where the canonicalized zone file will be written to.
        input_format (str, optional): The format in which the zone information is currently
                                      presented in the zone file.
                                      This should be either 'text' or 'raw'. Defaults to 'text'.

    Returns:
        list: A list of strings, each string being a line from the output of the named-compilezone
            command that indicates an error which prevented the DNS zone from being loaded.
            If there are no such errors, an empty list is returned.

    Raises:
        ValueError: If `input_format` is neither 'text' or 'raw'

    Usage:
        Call this function to check for, log, and return errors in a DNS zone file as it is being
        processed:
        >>> errors = named_compilezone('my_zone', '/path/to/my_zone_file', '/path/to/output_file')
        >>> if errors:
        ...     print('Errors found during zone processing.')
        ...     for error in errors:
        ...         print(error)
    """

    if input_format not in ["raw", "text"]:
        raise ValueError('specify one of "text" or "raw" value')
    command = (
        f"named-compilezone -k ignore -i none -f {input_format} -o {output_file} "
        f"{zone_name} {zone_file}"
    )
    logging.debug(command)
    try:
        subprocess.run(
            [
                "named-compilezone",
                "-i", "none",
                "-k", "ignore",
                "-m", "ignore",
                "-n", "ignore",
                "-r", "ignore",
                "-s", "full",
                "-f", input_format,
                "-o", output_file,
                zone_name,
                zone_file,
            ],
            check=True,
            capture_output=True,
            text=True,
        )
    except subprocess.CalledProcessError as err:
        if (
            int(err.returncode) > 0
            and "not loaded due to errors." in err.stdout
        ):
            logging.error("err_code: %s", err.returncode)
            for line in err.stdout.split("\n"):
                if line:
                    logging.error(line)
            return _parse_named_checkzone_log(err.stdout)

        for line in err.stdout.split("\n"):
            if line:
                logging.debug(line)
            return []
    return []


rewrite_zone_file = named_compilezone


def remove_lines_from_file(
    file_path: str, lines_to_remove: list, output_path: str = None
) -> None:
    """
    The function remove_lines_from_file removes specific lines from a file.

    Args:
        file_path (str): The fully qualified path to the file from which lines are to be removed.
        lines_to_remove (list): A list of integers, with each integer being a line
                                number (1-indexed) of the line to be removed from the file.
        output_path (str, optional): Path to the output file. If provided, the function will
                                     write result to this file.
                                     If not provided, the function will overwrite the original
                                     file. Defaults to None.

    Returns:
        None

    Logging:
        This function logs warning messages for each line that it removes from the file.
        It also logs an info message once the file has been successfully rewritten.

    Usage:
        Use this function to remove specific lines from a file:
        >>> remove_lines_from_file('/path/to/my_file', [2, 3])
        This will remove the 2nd and 3rd lines from the file located at /path/to/my_file.

        To write the result to a different file, specify the output_path argument:
        >>> remove_lines_from_file('/path/to/my_file', [2, 3], '/path/to/output_file')
        This will remove the 2nd and 3rd lines from the file located at /path/to/my_file
        and write the result to a file located at /path/to/output_file.
    """

    if not output_path:
        output_path = file_path
    logging.warning("file: %s lines to remove: %s", file_path, lines_to_remove)
    with open(file_path, "r", encoding="utf8") as fin:
        lines = fin.readlines()
        ptr = 1
        with open(output_path, "w", encoding="utf8") as fout:
            for line in lines:
                if ptr not in lines_to_remove:
                    fout.write(line)
                else:
                    logging.warning(
                        "file: %s - removing line: %s",
                        output_path,
                        line.strip(),
                    )
                ptr += 1
            logging.info("file %s rewritten", output_path)


def _parse_named_checkzone_log(error_output: str) -> list:
    """
    post-process the error output from named-checkzone and return offending lines

    :param error_output: err.stdout from running named-checkzone on rewriting a zone

    :return: list of the offending line numbers
    """
    line_numbers = []
    error_conditions = [
        "dns_rdata_fromtext",
        "dns_master_load",
    ]
    for line in error_output.split("\n"):
        parts = line.split(":")
        if any(error_condition in line for error_condition in error_conditions):
            line_numbers.append(int(parts[2]))
        if "NS record " in line and " appears to be an address" in line:
            line_numbers.append(int(parts[1]))
    return line_numbers


def get_csv_header(csvfile: io.TextIOWrapper) -> list:
    """
        The function get_csv_header retrieves the header from a CSV file.

        Args:
            csvfile (io.TextIOWrapper): A file object for the CSV file from which to retrieve the
            header.

        Returns:
            list: A list of strings where each string is a column name from the CSV file's header row.
                If the CSV file has no header row, an empty list is returned.

        Usage:
            Use this function to get the header from a CSV file:
            >>> with open('/path/to/my_file.csv', 'r') as csv_file:
            ...     # noinspection PyTypeChecker
    header = get_csv_header(csv_file)
            ...     print(header)
    """
    csvfile.seek(0)
    sample = csvfile.read(1024)
    dialect = csv.Sniffer().sniff(sample)
    cols = []
    if csv.Sniffer().has_header(sample):
        csvfile.seek(0)
        myreader = csv.reader(csvfile, dialect)
        cols = next(myreader)
    return cols


def csv_filtered_header(row: dict, col_filter: list = None) -> dict:
    """
    fetches csv required fields for header w/ any other(s) in filter list

    @param row: csv header row with fieldnames
    @param col_filter: list of other columns to filter on
    @return header_object: dictionary of column and index in the header row
    """
    header_object = {}
    items = list(row.keys())
    for col in row:
        if "header-" in col:
            idx = items.index(col)
            header_object[col] = idx
        elif col.endswith("*"):
            idx = items.index(col)
            header_object[col] = idx
        elif col_filter and col in col_filter:
            idx = items.index(col)
            header_object[col] = idx
    logging.debug(header_object)
    return header_object


def extract_filename_from_url(url) -> str:
    """
    The function extract_filename_from_url retrieves the name of the CSV file from a given URL.

    Args:
        url (str): An Infoblox download URL for CSV export.

    Returns:
        str: The name of the file that would be downloaded from the specified URL.

    Raises:
        Exception: An exception is raised if the URL cannot be parsed.

    Logging:
        This function logs debug messages as it parses the URL, and an error message if an exception
        was raised while parsing.

    Usage:
        Use this function to retrieve the name of a file from an Infoblox download URL:
        >>> file_name = extract_filename_from_url('https://my-infoblox-instance/export.csv')
    """

    logging.debug("parsing url: %s", url)
    try:
        obj = urlparse(url)
    except Exception as err:
        logging.error(err)
        raise
    logging.debug(pprint.pformat(obj))
    parts = obj.path.split("/")
    filename = str(parts[-1]).lower()
    logging.debug("parsed filename %s", filename)

    return filename


def ibx_csv_file_split(filename: str, output_path: str = "."):
    """
    The function ibx_csv_file_split splits a globally exported CSV file into separate CSV files
    based on the CSV object type(s).

    Args:
        filename (str): The name of the source CSV file to be split.
        output_path (str, optional): The directory where the output CSV files will be written.
            If the directory does not exist, it will be created. Defaults to the current directory.

    Raises:
        Exception: An exception is raised if the output directory cannot be created.

    Logging:
        This function logs warning messages when a CSV object type has no associated objects in
        the source CSV.
        It also logs informational messages when it creates output CSV files.

    Usage:
        Use this function to split a `Infoblox` exported CSV file into separate CSV files for
        each CSV object type:
        >>> ibx_csv_file_split('/path/to/my_exported_csv_file', '/path/to/output_directory')
    """

    if not os.path.exists(output_path):
        logging.warning(
            "output path %s does not exist, creating...", output_path
        )
        try:
            os.makedirs(output_path, exist_ok=True)
        except Exception as err:
            logging.error(err)
            raise

    # build the struct using the csv object type as the key
    csv_objects = {}
    with open(filename, "r", encoding="utf8") as handle:
        myreader = csv.reader(handle)
        for row in myreader:
            if row[0].lower().startswith("header-"):
                obj_type = row[0].lower().replace("header-", "")
                csv_output_file = f"{obj_type}.csv"
                csv_output_file_path = os.path.join(
                    output_path, csv_output_file
                )
                csv_objects[obj_type] = {}
                csv_objects[obj_type]["filename"] = csv_output_file
                csv_objects[obj_type]["filepath"] = csv_output_file_path
                csv_objects[obj_type]["data"] = [row]
            else:
                obj_type = row[0].lower()
                csv_objects[obj_type]["data"].append(row)

    # output the struct
    for obj_type, obj_data in csv_objects.items():
        num_of_objects = len(obj_data["data"])
        if num_of_objects == 1:
            logging.warning(
                "skipping creating output file %s, no %s objects",
                obj_data["filename"],
                obj_type,
            )
        else:
            logging.info(
                "creating output file %s with %s %s objects",
                obj_data["filename"],
                num_of_objects,
                obj_type,
            )
            with open(obj_data["filepath"], "w", encoding="utf8") as fh_out:
                mywriter = csv.writer(fh_out)
                for row in obj_data["data"]:
                    mywriter.writerow(row)
                fh_out.close()


def _get_include_data(chroot: str, include_file: str):
    data = ""
    full_path = f"{chroot}{include_file}"
    if not os.path.exists(full_path):
        logging.error("%s include file does not exist!", full_path)
        return data
    with open(full_path, "r", encoding="utf8") as file:
        for line in file:
            if "include " in line:
                logging.warning("nested include file found in %s", include_file)
                _, _inc_file, _ = line.split('"', 2)
                content = _get_include_data(chroot, _inc_file)
                if content:
                    data += content
            else:
                data += line
        return data


def generate_from_includes(chroot: str, filepath: str) -> str:
    """
    The function generate_from_includes generates a configuration file from include(s) directives
    found in another configuration file.

    Args:
        chroot (str): The path to the chroot environment where the config file to process exists.
        filepath (str): Path to the initial config file to process. The path is relative to
                        chroot if
                            it starts with '/'.

    Returns:
        str: A single string containing the full contents of the config file. This includes data
        from
            the initial config file and all files included with include directives. If the file
            specified
            in the filepath parameter does not exist, an empty string is returned.

    Logging:
        This function logs informational messages indicating the files it's processing and any
        'include'
        directives it encounters. An error message is logged if it attempts to process a file
        that doesn't exist.

    Usage:
        Use this function to generate a full config file from one that includes other files with
        include directives:
        >>> full_config = generate_from_includes('/path/to/chroot', '/path/to/my_config')
    """

    if filepath.startswith("/"):
        filepath = filepath.lstrip("/")
    logging.info("processing file %s", f"{chroot}{filepath}")
    data = ""
    if os.path.exists(os.path.join(chroot, filepath)):
        with open(os.path.join(chroot, filepath), "r", encoding="utf8") as file:
            for line in file:
                if line.startswith("include "):
                    logging.debug(line.strip())
                    _, include_file, _ = line.split('"', 2)
                    logging.info("processing include file %s", include_file)
                    contents = _get_include_data(chroot, include_file)
                    if contents:
                        data += contents
                else:
                    data += line
    else:
        logging.error("file %s does not exist!", os.path.join(chroot, filepath))
    return data



"""
Copyright 2023 Infoblox

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import os
import pprint
import re
from typing import Literal, Optional

import httpx

from ibx_sdk.nios.exceptions import WapiRequestException
from ibx_sdk.util import util

CsvOperation = Literal[
    "INSERT",
    "UPDATE",
    "DELETE",
    "REPLACE",
    "MERGE",
    "OVERRIDE",
    "CUSTOM",
]
GridRestoreMode = Literal["NORMAL", "FORCED", "CLONE"]
SupportedAlgorithms = Literal["SHA-1", "SHA-256", "SHA-384", "SHA-512"]
SupportedKeySizes = Literal[1024, 2048, 4096]
SupportedCertUsages = Literal[
    "ADMIN", "CAPTIVE_PORTAL", "SFNT_CLIENT_CERT", "IFMAP_DHCP"
]
SupportedCertTypes = Literal[
    "ADMIN",
    "CAPTIVE_PORTAL",
    "SFNT_CLIENT_CERT",
    "IFMAP_DHCP",
    "EAP_CA",
    "TAE_CA",
]
LogType = Literal[
    "SYSLOG",
    "AUDITLOG",
    "MSMGMTLOG",
    "DELTALOG",
    "OUTBOUND",
    "PTOPLOG",
    "DISCOVERY_CSV_ERRLOG",
]
MemberDataType = Literal[
    "DNS_CACHE",
    "DNS_CFG",
    "DHCP_CFG",
    "DHCPV6_CFG",
    "TRAFFIC_CAPTURE_FILE",
    "DNS_STATS",
    "DNS_RECURSING_CACHE",
]


class NiosFileopMixin:
    """
    NiosFileopMixin class
    """

    async def csv_export(
        self, wapi_object: str, filename: Optional[str] = None
    ) -> None:
        """
        Performs a CSV export for the provided WAPI object(s) and downloads the resulting file.

        This method sends a request to export data as CSV for the given WAPI object(s),
        retrieving the download URL and associated token upon success.

        It then downloads the CSV data file and notifies the server once the download is complete.

        Args:
            wapi_object (str): The WAPI object(s) to export.
            filename (Optional[str]): The optional file path for saving the exported CSV.
                If provided, hyphens in the filename are replaced with underscores.

        Raises:
            WapiRequestException: If a request-related error occurs during exporting or
                downloading the CSV file.

        """
        if filename:
            (_, filename) = os.path.split(filename)
            filename = os.path.join(_, filename.replace("-", "_"))

        # Call WAPI fileop  csv_export function
        logging.info("performing csv export for %s object(s)", wapi_object)
        payload = {"_object": wapi_object}
        try:
            response = await self.post(
                "fileop",
                params={"_function": "csv_export"},
                json=payload,
            )
            logging.debug(response.text)
            response.raise_for_status()
            try:
                obj = response.json()
            except httpx.DecodingError as exc:
                logging.error(f"DecodingError: {exc}")
                raise WapiRequestException(response.text) from exc
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(exc) from exc

        download_url = obj.get("url")
        download_token = obj.get("token")

        logging.info("downloading data from %s", download_url)

        if not filename:
            filename = util.extract_filename_from_url(download_url)

        await self.__download_file(download_url, filename)

        await self.__download_complete(download_token, filename)

    async def file_download(
        self,
        token: str,
        url: str,
        filename: str = None,
    ) -> None:
        """
        file_download downloads the generated file from the NIOS Grid using a token and url

        Args:
            token: Authentication token required for the download completion.
            url: URL of the file to be downloaded.
            filename: Optional; name for the downloaded file. If not provided, it will be extracted
            from the URL.

        Returns:
            None
        """
        logging.info("downloading data from %s", url)
        if not filename:
            filename = util.extract_filename_from_url(url)

        try:
            await self.__download_file(url, filename)
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(exc) from exc

        try:
            await self.__download_complete(token, filename)
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(exc) from exc

    async def file_upload(self, filename: str) -> str:
        """
        Perform a file upload into the NIOS Grid.

        Args:
            filename: The path of the file to be uploaded.

        Returns:
            str: The token received upon successful upload initialization.

        Raises:
            WapiRequestException: If there is a request exception during the upload process.
        """
        (path, filename) = os.path.split(filename)
        valid_filename = filename.replace("-", "_")

        # Call WAPI fileop Upload INIT
        logging.info("step 1 - request uploadinit %s", filename)
        try:
            obj = await self.__upload_init(filename=valid_filename)
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(exc) from exc

        upload_url = obj.get("url")
        token = obj.get("token")

        # specify a file handle for the file data to be uploaded
        with open(os.path.join(path, filename), "rb") as fh:
            fh.seek(0)
            upload_file = {"file": fh.read()}

            # Upload the contents of the CSV file
            logging.info(
                "step 2 - post the files using the upload_url provided"
            )
            try:
                await self.__upload_file(upload_url, upload_file)
            except httpx.TimeoutException as exc:
                logging.error(f"Timeout error: {exc}")
                raise WapiRequestException(exc) from exc
            except httpx.HTTPStatusError as exc:
                logging.error(f"HTTP error: {exc}")
                raise WapiRequestException(exc) from exc
            except httpx.RequestError as exc:
                logging.error(f"Request error: {exc}")
                raise WapiRequestException(exc) from exc
            else:
                return token

    async def upload_certificate(
        self,
        member: str,
        filename: str,
        certificate_usage: SupportedCertTypes = "ADMIN",
    ):
        """
        Upload an SSL Certificate file to the Grid

        Args:
            member: The member identifier to which the certificate will be uploaded.
            filename: The filename of the certificate to be uploaded.
            certificate_usage: The usage type of the certificate.
            The default is "ADMIN".

        Raises:
            WapiRequestException:
            If there is an error during the request to upload the certificate.
        """
        token = self.file_upload(filename=filename)

        # submit the task to CSV Job Manager
        logging.info(
            "step 3 - upload %s certificate on %s",
            certificate_usage,
            member,
        )
        payload = {
            "certificate_usage": certificate_usage,
            "member": member,
            "token": token,
        }
        try:
            res = self.post(
                "fileop",
                params={"_function": "uploadcertificate"},
                json=payload,
            )
            logging.debug(pprint.pformat(res.text))
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

    async def csv_import(
        self,
        task_operation: CsvOperation,
        csv_import_file: str,
        exit_on_error: bool = False,
    ) -> dict:
        """
        Perform a CSV import task using the NIOS CSV Task Manager

        Args:
            task_operation (CsvOperation): The operation to be performed on the CSV file. Should be
                                           a value from the `CsvOperation` enum.
            csv_import_file (str): The path to the CSV file to be imported.
            exit_on_error (bool): Indicates whether the program should exit if an error occurs
                                  during the import process. The default value is `False`.

        Returns:
            A dictionary containing the result of the CSV import task.

        Raises:
            httpx.RequestError: If an error occurs while making HTTP requests.
        """
        token = await self.file_upload(filename=csv_import_file)

        # submit the task to CSV Job Manager
        logging.info(
            "step 3 - execute the csv_import %s job on %s",
            task_operation,
            csv_import_file,
        )
        try:
            csvtask = await self.__csv_import(
                task_operation.upper(), token, exit_on_error
            )
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)
        else:
            return csvtask

    async def csvtask_status(self, csvtask: dict) -> dict:
        """
        Fetch the status of a CSV submitted task

        Args:
            csvtask (dict): The dictionary containing the information about the CSV import task.

        Returns:
            dict: The JSON response containing the status of the CSV import task.

        Raises:
            RequestException: If there is an error in the request.

        Description:
            This method is used to check the status of a CSV import task. It takes a dictionary,
            `csvtask`, as input, which contains the information about the CSV import task. The
            `csvtask` dictionary should have the following structure:

            {
                'csv_import_task': {
                    '_ref': '<CSV import task reference>'
                }
            }

            The method first retrieves the `_ref` value from the `csvtask` dictionary to identify
            the import task. It then sends a request to retrieve the status of the import task
            using the `get` method.

            If the status request is successful, the JSON response containing the status is
            logged using the `debug` level.

            Finally, the method returns the JSON response containing the status of the CSV import
            task.

            If there is any error in the request, a `RequestException` is raised and logged using
            the `error` level.

        Example usage:

        ```python
        csvtask = {
            'csv_import_task': {
                '_ref': '12345678'
            }
        }
        status = csvtask_status(csvtask)
        ```
        """
        _ref = csvtask["csv_import_task"]["_ref"]
        logging.debug("Checking status of csvimporttask %s", _ref)
        try:
            res = self.get(_ref)
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)
        else:
            logging.debug(res.json())

        return res.json()

    async def get_csv_errors_file(self, filename: str, job_id: str) -> None:
        """
        Fetches the csv-errors file for a specific job ID.

        Args:
            filename (str): The filename to be used when saving the csv-errors file.
            job_id (str): The job ID for which the csv-errors file should be fetched.

        Returns:
            None

        Raises:
            httpx.RequestError: If there is an error during the request.

        """
        logging.debug("fetching csv-errors file for job id %s", job_id)
        payload = {"import_id": job_id}
        try:
            res = await self.post(
                "fileop",
                params={"_function": "csv_error_log"},
                json=payload,
            )
            logging.debug(pprint.pformat(res.text))
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

        obj = res.json()
        token = obj.get("token")
        download_url = obj.get("url")

        csv_error_file = f"csv-errors-{filename}.csv"
        try:
            await self.__download_file(download_url, csv_error_file)
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

        # We're done - so post to downloadcomplete function
        try:
            await self.__download_complete(token, csv_error_file)
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

    async def download_certificate(
        self,
        member: str,
        certificate_usage: SupportedCertTypes = "ADMIN",
    ):
        """
        Download the SSL certificate from the Grid.

        Args:
            member: The identifier of the member for whom the certificate is being downloaded.
            certificate_usage: The type of certificate to be downloaded (e.g., "ADMIN").
        """
        logging.info(
            "Downloading %s certificate for %s",
            certificate_usage,
            member,
        )
        payload = {
            "member": member,
            "certificate_usage": certificate_usage,
        }
        logging.debug("json payload %s", payload)

        try:
            res = await self.post(
                "fileop",
                params={"_function": "downloadcertificate"},
                json=payload,
            )
            logging.debug(res.text)
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

        obj = res.json()
        download_url = obj.get("url")
        download_token = obj.get("token")

        await self.file_download(token=download_token, url=download_url)

    async def generate_selfsigned_cert(
        self,
        cn: str,
        member: str,
        days_valid: int = 365,
        algorithm: SupportedAlgorithms = "SHA-256",
        certificate_usage: SupportedCertUsages = "ADMIN",
        comment: Optional[str] = None,
        country: Optional[str] = None,
        email: Optional[str] = None,
        key_size: Optional[SupportedKeySizes] = 2048,
        locality: Optional[str] = None,
        org: Optional[str] = None,
        org_unit: Optional[str] = None,
        state: Optional[str] = None,
        subject_alternative_names: Optional[list[dict[str, str]]] = None,
    ):
        """
        Generate a Self-Signed Certificate on the Grid.

        Args:
            cn: The common name for the certificate.
            member: The member name for the certificate.
            days_valid: The number of days the certificate is valid for. Default is 365.
            algorithm: The algorithm used for certificate generation. The default is "SHA-256".
            certificate_usage: The usage type of the certificate. The default is "ADMIN".
            comment: Optional comment associated with the certificate.
            country: Optional country code.
            email: Optional email address.
            key_size: The size of the key used in certificate generation. Default is 2048.
            locality: Optional locality (e.g., city).
            org: Optional organization name.
            org_unit: Optional organizational unit.
            state: Optional state or province.
            subject_alternative_names: Optional list of subject alternative names.

        """
        logging.info("generating self-signed certificate for %s", member)
        payload: dict[str, str | int | list[dict[str, str]]] = {
            "cn": cn,
            "member": member,
            "algorithm": algorithm,
            "certificate_usage": certificate_usage,
            "days_valid": days_valid,
        }
        if comment:
            payload["comment"] = comment
        if country:
            payload["country"] = country
        if email:
            payload["email"] = email
        if key_size:
            payload["key_size"] = key_size
        if locality:
            payload["locality"] = locality
        if org:
            payload["org"] = org
        if org_unit:
            payload["org_unit"] = org_unit
        if state:
            payload["state"] = state
        if subject_alternative_names:
            payload["subject_alternative_names"] = subject_alternative_names
        logging.debug("json payload %s", payload)

        try:
            res = await self.post(
                "fileop",
                params={"_function": "generateselfsignedcert"},
                json=payload,
            )
            logging.debug(res.text)
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

        obj = res.json()
        download_url = obj.get("url")
        download_token = obj.get("token")

        await self.file_download(token=download_token, url=download_url)

    async def generate_csr(
        self,
        cn: str,
        member: str,
        algorithm: SupportedAlgorithms = "SHA-256",
        certificate_usage: SupportedCertUsages = "ADMIN",
        comment: Optional[str] = None,
        country: Optional[str] = None,
        email: Optional[str] = None,
        key_size: Optional[SupportedKeySizes] = 2048,
        locality: Optional[str] = None,
        org: Optional[str] = None,
        org_unit: Optional[str] = None,
        state: Optional[str] = None,
        subject_alternative_names: Optional[list[dict]] = None,
    ) -> None:
        """
        Generate a Certificate Signing Request

        Generate and download a CSR for a member of the grid.
        Once the CSR is generated, it is downloaded and saved locally to the
        current working directory.

        Args:
            cn: Common Name for the certificate.
            member: The member for which the certificate is being generated.
            algorithm: Algorithm used for certificate generation, default is "SHA-256".
            certificate_usage: Purpose of the certificate, default is "ADMIN".
            comment: Optional comment for the certificate.
            country: Optional country code for the certificate.
            email: Optional email address for the certificate.
            key_size: Optional key size for the certificate, default is 2048.
            locality: Optional locality or city for the certificate.
            org: Optional organization name for the certificate.
            org_unit: Optional organizational unit for the certificate.
            state: Optional state or province for the certificate.
            subject_alternative_names:
            Optional list of subject alternative names (SANs) for the
                    certificate.

        Returns:
            None
        """
        logging.info(locals())
        logging.info("generating %s csr for %s", certificate_usage, member)
        payload = {
            "cn": cn,
            "member": member,
            "algorithm": algorithm,
            "certificate_usage": certificate_usage,
        }
        # optional params
        for param in [
            "country",
            "email",
            "key_size",
            "comment",
            "locality",
            "org",
            "org_unit",
            "state",
            "subject_alternative_names",
        ]:
            if locals()[param] is not None:
                payload[param] = locals()[param]

        logging.debug("json payload %s", payload)

        try:
            res = await self.post(
                "fileop",
                params={"_function": "generatecsr"},
                json=payload,
            )
            logging.debug(res.text)
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

        obj = res.json()
        download_url = obj.get("url")
        download_token = obj.get("token")

        await self.file_download(token=download_token, url=download_url)

    async def get_log_files(
        self,
        log_type: LogType,
        filename: Optional[str] = None,
        endpoint: Optional[str] = None,
        include_rotated: bool = False,
        member: Optional[str] = None,
        msserver: Optional[str] = None,
        node_type: Optional[Literal["ACTIVE", "BACKUP"]] = None,
    ):
        """
        Fetch the log files for the provided member or msserver

        Args:
            log_type (LogType): The type of log files to fetch.
            filename (str): The name of the log file to download (Default value = None)
            endpoint (str): The specific endpoint for which to fetch log files. (Default: None)
            include_rotated (bool): Whether to include rotated log files. (Default: False)
            member (str): The member for which to fetch log files. (Default: None)
            msserver (str): The msserver for which to fetch log files. (Default: None)
            node_type: The type of node for which to fetch log files. Can be 'ACTIVE' or
                       'BACKUP'. (Default: None)
        """
        logging.info("fetching %s log files for %s", log_type, member)
        payload = {
            "log_type": log_type,
            "include_rotated": include_rotated,
        }

        if endpoint:
            payload["endpoint"] = endpoint
        if member:
            payload["member"] = member
        if node_type:
            payload["node_type"] = node_type
        if msserver:
            payload["msserver"] = msserver

        logging.debug("json payload %s", payload)

        try:
            res = await self.post(
                "fileop",
                params={"_function": "get_log_files"},
                json=payload,
            )
            logging.debug(res.text)
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

        obj = res.json()
        download_url = obj.get("url")
        download_token = obj.get("token")

        await self.file_download(
            token=download_token, url=download_url, filename=filename
        )

    async def get_support_bundle(
        self,
        member: str,
        filename: Optional[str] = None,
        cached_zone_data: bool = False,
        core_files: bool = False,
        log_files: bool = False,
        nm_snmp_logs: bool = False,
        recursive_cache_file: bool = False,
        remote_url: Optional[str] = None,
        rotate_log_files: bool = False,
    ):
        """
        Get the support bundle for a member.

        Args:
            member (str): The member for which to retrieve the support bundle.
            filename (str, optional): The filename of the support bundle.
            cached_zone_data (bool, optional): Whether to include cached zone data in the support
                                               bundle. Defaults to False.
            core_files (bool, optional): Whether to include core files in the support bundle.
                                         Defaults to False.
            log_files (bool, optional): Whether to include log files in the support bundle.
                                        Defaults to False.
            nm_snmp_logs (bool, optional): Whether to include NM SNMP logs in the support bundle.
                                           Defaults to False.
            recursive_cache_file (bool, optional): Whether to include a recursive cache file in the
                                                   support bundle. Defaults to False.
            remote_url (str, optional): The remote URL where the support bundle will be uploaded.
                                        Defaults to None.
            rotate_log_files (bool, optional): Whether to rotate log files before creating the
                                               support bundle. Defaults to False.

        Raises:
            httpx.RequestError: If an error occurs during the request.

        """
        logging.info("performing get_support_bundle for %s object(s)", member)
        payload = {
            "member": member,
            "cached_zone_data": cached_zone_data,
            "core_files": core_files,
            "log_files": log_files,
            "nm_snmp_logs": nm_snmp_logs,
            "recursive_cache_file": recursive_cache_file,
            "rotate_log_files": rotate_log_files,
        }
        if remote_url:
            payload["remote_url"] = remote_url
        logging.debug(pprint.pformat(payload))
        try:
            res = await self.post(
                "fileop",
                params={"_function": "get_support_bundle"},
                json=payload,
                timeout=None,
            )
            logging.debug(res.text)
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

        obj = res.json()
        download_url = obj.get("url")
        download_token = obj.get("token")

        await self.file_download(
            token=download_token, url=download_url, filename=filename
        )

    async def grid_backup(self, filename: Optional[str] = None) -> None:
        """
        Perform a NIOS Grid Backup.

        Args:
            filename: str, optional. The name of the backup file. The default is 'database.bak'.

        Returns:
            None

        Raises:
            httpx.RequestError: If an error occurs during the backup process.
        """
        payload = {"type": "BACKUP"}

        logging.info("step 1 - request gridbackup %s", filename)
        try:
            res = await self.__getgriddata(payload)
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

        token = res.get("token")
        download_url = res.get("url")

        logging.info("step 2 - saving backup to %s", filename)
        await self.file_download(
            token=token, url=download_url, filename=filename
        )

    async def grid_restore(
        self,
        filename: str = "database.bak",
        mode: GridRestoreMode = "NORMAL",
        keep_grid_ip: bool = False,
    ):
        """
        Perform a NIOS Grid restore of a database using a given file.

        Args:
            filename (str): The filename of the database file to be restored.
                The default is "database.bak".
            mode (GridRestoreMode): The restore mode to be used. Default is "NORMAL".
            keep_grid_ip (bool): Indicates whether to keep the grid IP address. Default is False.

        """
        token = await self.file_upload(filename=filename)

        # Execute the restore
        logging.info("step 3 - execute the grid restore")
        try:
            await self.__restore_database(keep_grid_ip, mode, token)
        except httpx.RequestError as exc:
            logging.error("step 3 - Error: %s", exc)
            raise WapiRequestException(exc)
        logging.info("Grid restore successful!")

    async def member_config(
        self,
        member: str,
        conf_type: MemberDataType,
        filename: Optional[str] = None,
        remote_url: str = None,
    ) -> None:
        """
        Fetch member configuration file for a given service type.

        Args:
            member: A string representing the grid member.
            conf_type: An enum representing the type of the config file.
            filename: A string value of the filename to save
            remote_url: An optional string representing the remote URL.

        Returns:
            A string representing the downloaded file.

        """
        conf_type = conf_type.upper()
        logging.info(
            "fetching %s config file for grid member %s",
            conf_type,
            member,
        )
        payload = {"member": member, "type": conf_type}
        if remote_url:
            payload["remote_url"] = remote_url
        try:
            res = await self.post(
                "fileop",
                params={"_function": "getmemberdata"},
                json=payload,
            )
            logging.debug(res.text)
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

        obj = res.json()
        download_url = obj.get("url")
        download_token = obj.get("token")

        await self.file_download(
            token=download_token, url=download_url, filename=filename
        )

    async def get_lease_history(
        self,
        member: str,
        start_time: int = None,
        end_time: int = None,
        remove_url: str = None,
    ) -> None:
        """
        fetch DHCP lease history files from a NIOS Grid Member

        Args:
            member: A string representing the grid member that the DHCP lease history is being fetched from.
            start_time: An optional integer representing the start time in epoch format. Defaults to None.
            end_time: An optional integer representing the end time in epoch format. Defaults to None.
            remove_url: An optional string representing the remove URL. Defaults to None.

        Returns:
            A string representing the filename of the downloaded DHCP lease history file.

        Raises:
            WapiRequestException: If there is an error in the API request.

        """
        logging.info("fetching DHCP lease history from grid member %s", member)
        payload: dict[str, str | int] = {"member": member}
        if start_time is not None:
            payload["start_time"] = start_time
        if end_time is not None:
            payload["end_time"] = end_time
        if remove_url is not None:
            payload["remove_url"] = remove_url
        try:
            res = await self.post(
                "fileop",
                params={"_function": "getleasehistoryfiles"},
                json=payload,
            )
            logging.debug(res.text)
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

        obj = res.json()
        download_url = obj.get("url")
        download_token = obj.get("token")

        await self.file_download(token=download_token, url=download_url)

    async def __csv_import(
        self,
        task_operation: str,
        upload_token: str,
        exit_on_error: bool = False,
    ) -> dict:
        headers = {"content-type": "application/json"}

        # set the request parameters
        payload = {
            "action": "START",
            "doimport": True,
            "on_error": "STOP" if exit_on_error else "CONTINUE",
            "operation": task_operation,
            "separator": "COMMA",
            "token": upload_token,
        }

        # Update the operation if the user passes in MERGE or OVERRIDE directly
        if task_operation == "MERGE":
            payload["operation"] = "UPDATE"
            payload["update_method"] = "MERGE"
        elif task_operation == "OVERRIDE":
            payload["operation"] = "UPDATE"
            payload["update_method"] = "OVERRIDE"

        # start the CSV task in job manager
        try:
            res = await self.post(
                "fileop",
                params={"_function": "csv_import"},
                json=payload,
                headers=headers,
                timeout=None,
            )
            logging.debug(pprint.pformat(res.text))
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

        return res.json()

    async def __download_complete(self, token: str, filename: str) -> None:
        header = {"Content-type": "application/json"}
        payload = {"token": token}
        try:
            res = await self.post(
                "fileop",
                params={"_function": "downloadcomplete"},
                json=payload,
                headers=header,
            )
            logging.info("file %s download complete", filename)
            res.raise_for_status()
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(exc) from exc

    async def __download_file(self, download_url, filename=None) -> None:
        download_url = await self.__update_url(url=download_url)
        header = {"Content-type": "application/force-download"}
        logging.info(download_url)
        self.conn.verify = self.ssl_verify
        async with self.conn.stream(
            "GET",
            download_url,
            headers=header,
        ) as res:
            res.raise_for_status()
            with open(filename, "wb") as file_out:
                async for chunk in res.aiter_bytes(chunk_size=1024):
                    file_out.write(chunk)

    async def __getgriddata(self, payload: dict) -> dict:
        headers = {"content-type": "application/json"}
        try:
            res = await self.post(
                "fileop",
                params={"_function": "getgriddata"},
                json=payload,
                headers=headers,
                timeout=None,
            )
            logging.debug(pprint.pformat(res.text))
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

        return res.json()

    async def __restore_database(
        self, keep_grid_ip: bool, mode: str, upload_token: str
    ) -> dict:
        # set the content type back to JSON
        headers = {"content-type": "application/json"}

        # set the request parameters
        payload = {
            "keep_grid_ip": keep_grid_ip,
            "mode": mode,
            "token": upload_token,
        }

        # start the restore
        try:
            res = await self.post(
                "fileop",
                params={"_function": "restoredatabase"},
                json=payload,
                headers=headers,
                timeout=None,
            )
            logging.debug(pprint.pformat(res.text))
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)
        return res

    async def __update_url(self, url: str) -> str:
        if self.grid_mgr in url:
            return url
        elif "[" in url and "]" in url:
            ipv6_pattern = r"https://(\[[a-zA-Z0-9:]+\])"
            return re.sub(ipv6_pattern, f"https://{self.grid_mgr}", url)
        else:
            ipv4_pattern = r"https://(\d{1,3}\.){3}\d{1,3}"
            return re.sub(ipv4_pattern, f"https://{self.grid_mgr}", url)

    async def __upload_file(self, upload_url: str, upload_file: dict) -> None:
        upload_url = await self.__update_url(upload_url)
        logging.debug(upload_url)
        try:
            res = await self.conn.post(
                upload_url,
                files=upload_file,
                timeout=None,
            )
            logging.debug(pprint.pformat(res.text))
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

    async def __upload_init(self, filename: str) -> dict:
        headers = {"content-type": "application/json"}
        payload = {"filename": filename}
        try:
            res = await self.post(
                "fileop",
                params={"_function": "uploadinit"},
                headers=headers,
                json=payload,
            )
            logging.debug(pprint.pformat(res.text))
            res.raise_for_status()
        except httpx.RequestError as exc:
            logging.error(exc)
            raise WapiRequestException(exc)

        return res.json()



"""
Copyright 2023 Infoblox

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import pprint
from typing import Literal, Optional

import httpx
from ibx_sdk.nios.exceptions import WapiRequestException


class NiosServiceMixin:
    """
    NIOS Service Mixin class
    """

    RestartMode = Literal["GROUPED", "SEQUENTIAL", "SIMULTANEOUS"]
    RestartOption = Literal["FORCE_RESTART", "RESTART_IF_NEEDED"]
    RestartServices = Literal["ALL", "DNS", "DHCP", "DHCPV4", "DHCPV6"]

    def service_restart(
        self,
        groups: Optional[list] = None,
        members: Optional[list[str]] = None,
        mode: Optional[RestartMode] = None,
        restart_option: Optional[RestartOption] = "RESTART_IF_NEEDED",
        services: Optional[list[RestartServices]] = None,
        user_name: Optional[str] = None,
    ) -> None:
        """
        Restarts specified services on the grid.

        Args:
            self (Gift): Gift object
            groups (Optional[list]): List of group names. Default is None.
            members (Optional[list[str]]): List of member names. Default is None.
            mode (Optional[ServiceRestartMode]): Restart mode. Default is None.
            restart_option (Optional[ServiceRestartOption]): Restart option. The default is
                            'RESTART_IF_NEEDED'.
            services (Optional[list[ServiceRestartServices]]): List of services to restart.
                                                               Default is None.
            user_name (Optional[str]): Username. Default is None.

        Returns:
            None

        Raises:
            httpx.RequestError: If there is an error, send the restart request.

        Examples:
            # Restart services in a group
            service_restart(groups=['group_name'])

            # Restart services of specific members
            service_restart(members=['member1', 'member2'])

            # Restart services with a specific restart mode
            service_restart(mode=ServiceRestartMode.IMMEDIATE)

            # Restart services with a specific restart option
            service_restart(restart_option=ServiceRestartOption.RESTART_IF_NEEDED)

            # Restart specific services
            service_restart(services=[ServiceRestartServices.SERVICE_NAME1,
            ServiceRestartServices.SERVICE_NAME2])

            # Restart services for a specific user
            service_restart(user_name='username')
        """
        data = {}
        if groups:
            data["groups"] = [groups]
        if members:
            data["members"] = [members]
        if mode:
            data["mode"] = mode
        if restart_option:
            data["restart_option"] = restart_option
        if services:
            data["services"] = [services]
        if user_name:
            data["user_name"] = user_name

        logging.debug(pprint.pformat(data))

        try:
            res = self.post(
                self.grid_ref,
                params={"_function": "restartservices"},
                json=data,
            )
            res.raise_for_status()
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(exc) from exc
        else:
            logging.info(
                "successfully restarted %s services", data.get("services")
            )

    def update_service_status(self, services: str = "ALL") -> None:
        """
        Updates the status of a service.

        Args:
            self (Gift): The instance of the class.
            services (str): The service option to update the status for.
                            The default value is 'ALL'.

        Returns:
            None

        Raises:
            httpx.RequestError: If an error occurs while making the request.
        """
        payload = {"service_option": services}
        try:
            res = self.post(
                self.grid_ref,
                params={"_function": "requestrestartservicestatus"},
                json=payload,
            )
            res.raise_for_status()
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(exc) from exc

    def get_service_restart_status(self) -> dict:
        """
        Retrieves the status of a service restart.

        Args:
            self (Gift): The instance of the Gift class

        Returns:
            dict: A dictionary containing the service restart status information.

        Raises:
            httpx.RequestError: If there is a general request error.
        """
        try:
            response = self.get("restartservicestatus")
            response.raise_for_status()
            try:
                return response.json()
            except httpx.DecodingError as exc:
                logging.error(f"DecodingError: {exc}")
                raise WapiRequestException(response.text) from exc
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(exc) from exc



"""
Copyright 2023 Infoblox

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import ssl
from typing import Union, Any, Optional, List

import httpx
import urllib3

from ibx_sdk.nios.exceptions import (
    WapiInvalidParameterException,
    WapiRequestException,
)
from ibx_sdk.nios.fileop import NiosFileopMixin
from ibx_sdk.nios.service import NiosServiceMixin

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class Gift(httpx.Client, NiosServiceMixin, NiosFileopMixin):
    """Handles interactions with the Infoblox WAPI.

    This class provides a range of classes to interact with Infoblox WAPI,
    including session management, data retrieval, file operations, and service management.

    Attributes:
        grid_mgr (str): IP address or hostname of the Grid Manager.
        wapi_ver (str): Version of the Infoblox WAPI.
        ssl_verify (bool): Flag to determine SSL certificate verification.
        conn (httpx.Client, optional): Active session to the WAPI grid. Default is None.
        grid_ref (str, optional): Reference ID of the connected grid. Default is None.

    Examples:

    Initialize the WAPI instance with a dictionary of properties:

    ```py

    wapi_properties = {
        'grid_mgr': 'gm.example.com',
        'wapi_ver': '2.11',
        'ssl_verify': False
    }
    wapi = Gift(wapi_properties)

    wapi.connect(username='admin', password='infoblox')

    ```

    Build up the WAPI instance one property at a time:

    ```python

    wapi = Gift()

    wapi.grid_mgr = 'gm.example.com'
    wapi.wapi_ver = '2.11'
    wapi.ssl_verify = False

    wapi.connect(username='admin', password='infoblox')

    ```
    """

    def __init__(
        self,
        grid_mgr: str = None,
        wapi_ver: str = "2.5",
        ssl_verify: bool | str = False,
        timeout: httpx.Timeout = 10.0,
    ) -> None:
        super().__init__()
        self.grid_mgr = grid_mgr
        self.wapi_ver = wapi_ver
        self.ssl_verify = ssl_verify
        self.timeout = timeout
        self.conn = None
        self.grid_ref = None

    def __repr__(self):
        args = []
        for key, value in self.__dict__.items():
            args.append(f"{key}={value}")
        return f"{self.__class__.__qualname__}({', '.join(args)})"

    @property
    def url(self) -> str:
        """
        Constructs a property using `grid_mgr` and `wapi_ver` attributes for the WAPI class.

        Parameters:
        - grid_mgr (str): The IP address or hostname of the grid manager.
        - wapi_ver (str): The version of the WAPI.

        Returns:
        - url (str): The URL constructed using the grid manager and WAPI version.

        Raises:
        - None.

        Example:

        ```python

        wapi = Gift()
        wapi.grid_mgr = '10.0.0.1'
        wapi.wapi_ver = '2.10'
        url = wapi.url

        print(url)

        ```
        The above code example will return output:

        ```
        'https://10.0.0.1/wapi/v2.10'
        ```

        """
        if self.grid_mgr and self.wapi_ver:
            return f"https://{self.grid_mgr}/wapi/v{self.wapi_ver}"
        return ""

    def connect(
        self,
        username: str = None,
        password: str = None,
        certificate: str = None,
    ) -> None:
        """
        Make a connection to the grid manager using the WAPI instance

        Args:
            username: A string representing the username for the connection. (default: None)
            password: A string representing the password for the connection. (default: None)
            certificate: A string representing the certificate for the connection. (default: None)
        Raises:
            WapiInvalidParameterException: If neither a username and password nor a certificate
                                           is provided.

        """
        if not self.url:
            logging.error("invalid url %s - unable to connect!", self.url)
            raise WapiInvalidParameterException

        if username and password:
            self.__basic_auth_request(username, password)
        elif certificate:
            self.__certificate_auth_request(certificate)
        else:
            raise WapiInvalidParameterException

    def __certificate_auth_request(self, certificate: str) -> dict | None:
        """
        This private method performs a certificate authentication request to the API. It uses the
        provided certificate to establish a connection with the API server using the httpx
        library.

        Args:
            certificate (str): The certificate to be used for authentication with the API.

        Returns:
            grid _ref (dict): A dictionary

        Raises:
            WapiRequestException: If there is an error with the request to the API.

        """
        ctx = ssl.create_default_context()
        ctx.load_cert_chain(certfile=certificate)
        if self.ssl_verify:
            ctx.load_verify_locations(cafile=self.ssl_verify)
        else:
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
        try:
            conn = httpx.Client(verify=ctx, timeout=self.timeout)
            res = conn.get(f"{self.url}/grid")
            res.raise_for_status()
            try:
                grid = res.json()
                self.conn = conn
                self.grid_ref = grid[0].get("_ref", "")
                return self.grid_ref
            except httpx.DecodingError as exc:
                logging.error(f"DecodingError: {res.text}")
                raise WapiRequestException(res.text) from exc
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(exc) from exc

    def __basic_auth_request(self, username: str, password: str) -> dict | None:
        """
        This private method makes a request to the specified URL with basic authentication using
        the provided username and password. It stores the session connection in the instance
        attribute 'conn*' and the grid reference in the instance attribute 'grid_ref'.

        Note:
            This method requires the 'httpx' library to be installed.

        Args:
            username (str): The username for basic authentication.
            password (str): The password for basic authentication.

        Returns:
            grid _ref (dict): A dictionary

        Raises:
            WapiRequestException: If an error occurs during the request.
        """
        auth = httpx.BasicAuth(username, password)

        ctx = ssl.create_default_context()
        if self.ssl_verify:
            ctx.load_verify_locations(cafile=self.ssl_verify)
        else:
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
        try:
            conn = httpx.Client(auth=auth, verify=ctx, timeout=self.timeout)
            res = conn.get(f"{self.url}/grid")
            res.raise_for_status()
            try:
                grid = res.json()
                self.conn = conn
                self.grid_ref = grid[0].get("_ref", "")
                return self.grid_ref
            except httpx.DecodingError as exc:
                logging.error(f"DecodingError: {exc}")
                raise WapiRequestException(res.text) from exc
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(exc) from exc

    def object_fields(self, wapi_object: str) -> Union[str, None]:
        """
        Retrieves the object fields for a specified WAPI object.

        Args:
            wapi_object (str): The name of the WAPI object for which to retrieve the fields.

        Returns:
            Union[str, None]: A string containing the fields separated by commas, or None if an
            error occurred.

        Raises:
            WapiRequestException: If there was an error connecting to the WAPI service.

        Example:

        ```py
        wapi = Gift()
        fields = wapi.object_fields('record:host')
        if fields is not None:
            print(f"Fields: {fields}")
        ```
        """
        try:
            logging.debug("trying %s/%s?_schema", self.url, wapi_object)
            res = self.conn.get(f"{self.url}/{wapi_object}?_schema")
            res.raise_for_status()
            try:
                data = res.json()
                fields = ",".join(
                    field["name"]
                    for field in data.get("fields")
                    if "r" in field.get("supports")
                )
                return fields
            except httpx.DecodingError as exc:
                logging.error(f"DecodingError: {exc}")
                raise WapiRequestException(res.text) from exc
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(exc) from exc

    def max_wapi_ver(self) -> None:
        """
        Retrieves and sets the maximum supported version of the WAPI.

        This method queries the WAPI to find the highest supported version and updates
        the `wapi_ver` attribute of the instance with this value. It does not return
        any value.

        Raises:
            WapiRequestException: If there is an error in making the GET request to retrieve
                                  the WAPI version. This could be due to network issues,
                                  authentication errors,
                                  or other problems in the request process.

        Example Usage:
        ```python
        session = Gift()
        session.max_wapi_ver()
        print(session.wapi_ver)  # Prints the maximum supported WAPI version
        ```

        Note:
            This method updates the `wapi_ver` attribute of the WAPI session instance.
        """

        url = f"https://{self.grid_mgr}/wapi/v1.0/?_schema"
        try:
            logging.debug("trying %s", url)
            res = self.conn.get(url)
            res.raise_for_status()
            try:
                data = res.json()
                versions = data.get("supported_versions")
                versions.sort(key=lambda s: list(map(int, s.split("."))))
                logging.debug(versions)
                max_wapi_ver = versions.pop()
                self.wapi_ver = max_wapi_ver
            except httpx.DecodingError as exc:
                logging.error(f"DecodingError: {exc}")
                raise WapiRequestException(res.text) from ValueError
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(exc) from exc

    def get(
        self, wapi_object: str, params: Optional[dict] = None, **kwargs: Any
    ) -> httpx.Response:
        """
        Return WAPI object(s).
        Args:
            wapi_object (str): The name of the WAPI object to retrieve.
            params (Optional[dict]): Optional parameters to include in the request URL.
            **kwargs: Additional keyword arguments to pass to the request.

        Returns:
            Response: The response object containing the result of the request.
        """
        url = f"{self.url}/{wapi_object}"
        res = None
        try:
            res = self.conn.get(url, params=params, **kwargs)
            res.raise_for_status()
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(res.text) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(res.text) from exc
        return res

    def getone(
        self, wapi_object: str, params: Optional[dict] = None, **kwargs: Any
    ) -> str:
        """
        Return the reference of a single WAPI object.

        Args:
            wapi_object: A string representing the object to retrieve data from.
            params: Optional dictionary of parameters to include in the request.
            **kwargs: Additional keyword arguments to be passed to the request.

        Returns:
            str: A string response containing the _ref of the WAPI object.

        Raises:
            WapiRequestException: If multiple data records were returned or no data was returned.
        """
        url = f"{self.url}/{wapi_object}"
        response = None
        try:
            response = self.conn.request("get", url, params=params, **kwargs)
            response.raise_for_status()
            try:
                data = response.json()
            except httpx.DecodingError as exc:
                logging.error(f"DecodingError: {response.text}")
                raise WapiRequestException(response.text) from exc
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(response.text) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(response.text) from exc
        else:
            if len(data) > 1:
                raise WapiRequestException(
                    "Multiple data records were returned"
                )
            elif len(data) == 0:
                raise WapiRequestException("No data was returned")
        return data[0].get("_ref", "")

    def post(
        self,
        wapi_object: str,
        data: Optional[Union[dict, str]] = None,
        json: Optional[dict] = None,
        **kwargs: Any,
    ) -> httpx.Response:
        """
        Create a POST request to create a WAPI object.

        Args:
            wapi_object: The object to which the POST request is being made.
            data (optional): The data to be sent in the body of the request.
                             Can be a dictionary or a string. Default is None.
            json (optional): The JSON data to be sent in the body of the request.
                             Should be a dictionary. Default is None.
            **kwargs (optional): Additional keyword arguments to be passed to the request.

        Returns:
            Response: The response object containing the server's response to the POST request.
        """
        url = f"{self.url}/{wapi_object}"
        res = None
        try:
            res = self.conn.request("post", url, data=data, json=json, **kwargs)
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(res.text) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(res.text) from exc
        return res

    def put(
        self,
        wapi_object_ref: str,
        data: Optional[Union[dict, str]] = None,
        **kwargs: Any,
    ) -> httpx.Response:
        """
        Create a PUT request to update a WAPI object by its _ref.

        Args:
            wapi_object_ref: The reference string for the WAPI object.
            data: Optional data to be sent with the request. Can be a dictionary or a string. Defaults to None.
            **kwargs: Additional keyword arguments to be passed to the request.

        Returns:
            Response: The response object for the PUT request.
        """
        url = f"{self.url}/{wapi_object_ref}"
        res = None
        try:
            res = self.conn.request("put", url, data=data, **kwargs)
            return res
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(res.text) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(res.text) from exc

    def delete(self, wapi_object_ref: str, **kwargs: Any) -> httpx.Response:
        """
        Args:
            wapi_object_ref (str): A string representing the reference to the WAPI object that
                                   needs to be deleted.
            **kwargs (Any): Additional keyword arguments that can be passed to the `request` method
                            of the connection object.

        Returns:
            A `Response` object representing the response received from the server.

        """
        url = f"{self.url}/{wapi_object_ref}"
        res = None
        try:
            res = self.conn.request("delete", url, **kwargs)
            return res
        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP error: {exc}")
            raise WapiRequestException(res.text) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error: {exc}")
            raise WapiRequestException(res.text) from exc

    def get_paginated(
        self,
        wapi_object: str,
        limit: int = 1000,
        params: Optional[dict] = None,
        **kwargs: Any
    ) -> List[dict]:
        """
        Fetches paginated data from the WAPI API.

        This method retrieves data in chunks using a pagination mechanism. The function
        handles API responses, manages pagination parameters, and consolidates the
        results into a list. It raises exceptions if timeout, HTTP status errors, or
        request errors are encountered during the process.

        Args:
            wapi_object: The name of the WAPI object to fetch data from.
            limit: Maximum number of records to retrieve per API request. Defaults to 1000.
            params: Additional query parameters to include in the request. Defaults to None.
            **kwargs: Additional keyword arguments passed to the HTTP GET request.

        Returns:
            A list of dictionaries containing the retrieved data from the WAPI API.

        Raises:
            WapiRequestException: If there is a timeout, HTTP status error, or
            request-related error during the API call.
        """
        results = []
        params = params.copy() if params else {}

        params.update({
            "_paging": 1,
            "_return_as_object": 1,
            "_max_results": limit,
        })

        url = f"{self.url}/{wapi_object}"
        response = None

        try:
            while True:
                response = self.conn.get(url, params=params, **kwargs)
                response.raise_for_status()

                data = response.json()
                results.extend(data.get("result", []))

                next_page_id = data.get("next_page_id")
                if not next_page_id:
                    break

                params["_page_id"] = next_page_id

        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error while fetching {url}: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP status error while fetching {url}: {exc}")
            raise WapiRequestException(response.text) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error while fetching {url}: {exc}")
            raise WapiRequestException(str(exc)) from exc

        return results


import logging
import ssl
from typing import Union, Optional, List, Any

import httpx

from ibx_sdk.nios.asynchronous.fileop import NiosFileopMixin
from ibx_sdk.nios.exceptions import (
    WapiInvalidParameterException,
    WapiRequestException,
)
from ibx_sdk.nios.service import NiosServiceMixin


class AsyncGift(httpx.AsyncClient, NiosServiceMixin, NiosFileopMixin):
    def __init__(
        self,
        grid_mgr: str = None,
        wapi_ver: str = "2.5",
        ssl_verify: Union[bool, str] = False,
    ):
        self.grid_mgr = grid_mgr
        self.wapi_ver = wapi_ver
        self.ssl_verify = ssl_verify
        self.conn = None
        self.grid_ref = None
        super().__init__()

    def __repr__(self):
        args = []
        for key, value in self.__dict__.items():
            args.append(f"{key}={value}")
        return f"{self.__class__.__qualname__}({', '.join(args)})"

    @property
    def url(self) -> str:
        return (
            f"https://{self.grid_mgr}/wapi/v{self.wapi_ver}"
            if self.grid_mgr and self.wapi_ver
            else ""
        )

    async def connect(
        self,
        username: str = None,
        password: str = None,
        certificate: str = None,
    ):
        if not self.url:
            logging.error("invalid url %s - unable to connect!", self.url)
            raise WapiInvalidParameterException

        if username and password:
            await self.__basic_auth_request(username, password)
        elif certificate:
            await self.__certificate_auth_request(certificate)
        else:
            raise WapiInvalidParameterException

    async def __basic_auth_request(self, username: str, password: str):
        auth = httpx.BasicAuth(username, password)
        ctx = ssl.create_default_context()
        if self.ssl_verify:
            ctx.load_verify_locations(cafile=self.ssl_verify)
        else:
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
        try:
            self.conn = httpx.AsyncClient(auth=auth, verify=ctx)
            res = await self.conn.get(f"{self.url}/grid")
            res.raise_for_status()
        except httpx.HTTPStatusError as exc:
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            raise WapiRequestException(exc) from exc
        else:
            grid = res.json()
            self.grid_ref = grid[0].get("_ref")

    async def __certificate_auth_request(self, certificate: str):
        ctx = ssl.create_default_context()
        ctx.load_cert_chain(certfile=certificate)
        if self.ssl_verify:
            ctx.load_verify_locations(cafile=self.ssl_verify)
        else:
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
        try:
            self.conn = httpx.AsyncClient(verify=ctx)
            res = await self.conn.get(f"{self.url}/grid")
            res.raise_for_status()
        except httpx.HTTPStatusError as exc:
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            raise WapiRequestException(exc) from exc
        else:
            grid = res.json()
            self.grid_ref = grid[0].get("_ref")

    async def get(
        self, wapi_object: str, params: Optional[dict] = None, **kwargs
    ) -> httpx.Response:
        res = await self.conn.get(
            f"{self.url}/{wapi_object}", params=params, **kwargs
        )
        res.raise_for_status()
        return res

    async def getone(
        self, wapi_object: str, params: Optional[dict] = None, **kwargs
    ) -> str:
        res = await self.conn.get(
            f"{self.url}/{wapi_object}", params=params, **kwargs
        )
        res.raise_for_status()
        data = res.json()
        if len(data) != 1:
            raise WapiRequestException("Expected exactly one result")
        return data[0].get("_ref", "")

    async def post(
        self, wapi_object: str, json: Optional[dict] = None, **kwargs
    ) -> httpx.Response:
        res = await self.conn.post(
            f"{self.url}/{wapi_object}", json=json, **kwargs
        )
        res.raise_for_status()
        return res

    async def put(
        self,
        wapi_object_ref: str,
        data: Optional[Union[dict, str]] = None,
        **kwargs,
    ) -> httpx.Response:
        res = await self.conn.put(
            f"{self.url}/{wapi_object_ref}", data=data, **kwargs
        )
        res.raise_for_status()
        return res

    async def delete(self, wapi_object_ref: str, **kwargs) -> httpx.Response:
        res = await self.conn.delete(f"{self.url}/{wapi_object_ref}", **kwargs)
        res.raise_for_status()
        return res

    async def object_fields(self, wapi_object: str) -> Union[str, None]:
        try:
            res = await self.conn.get(f"{self.url}/{wapi_object}?_schema")
            res.raise_for_status()
            data = res.json()
        except httpx.HTTPStatusError as exc:
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            raise WapiRequestException(exc) from exc
        else:
            fields = ",".join(
                field["name"]
                for field in data.get("fields")
                if "r" in field.get("supports")
            )
        return fields

    async def max_wapi_ver(self) -> None:
        url = f"https://{self.grid_mgr}/wapi/v1.0/?_schema"
        try:
            res = await self.conn.get(url)
            res.raise_for_status()
            data = res.json()
        except httpx.HTTPStatusError as exc:
            raise WapiRequestException(exc) from exc
        except httpx.RequestError as exc:
            raise WapiRequestException(exc) from exc
        else:
            versions = data.get("supported_versions")
            versions.sort(key=lambda s: list(map(int, s.split("."))))
            max_wapi_ver = versions.pop()
            setattr(self, "wapi_ver", max_wapi_ver)

    async def get_paginated(
            self,
            wapi_object: str,
            limit: int = 1000,
            params: Optional[dict] = None,
            **kwargs: Any
    ) -> List[dict]:
        """
        Fetches paginated data from the WAPI API.

        This method retrieves data in chunks using a pagination mechanism. The function
        handles API responses, manages pagination parameters, and consolidates the
        results into a list. It raises exceptions if timeout, HTTP status errors, or
        request errors are encountered during the process.

        Args:
            wapi_object: The name of the WAPI object to fetch data from.
            limit: Maximum number of records to retrieve per API request. Defaults to 1000.
            params: Additional query parameters to include in the request. Defaults to None.
            **kwargs: Additional keyword arguments passed to the HTTP GET request.

        Returns:
            A list of dictionaries containing the retrieved data from the WAPI API.

        Raises:
            WapiRequestException: If there is a timeout, HTTP status error, or
            request-related error during the API call.
        """
        results = []
        params = params.copy() if params else {}

        params.update({
            "_paging": 1,
            "_return_as_object": 1,
            "_max_results": limit,
        })

        url = f"{self.url}/{wapi_object}"
        response = None

        try:
            while True:
                response = await self.conn.get(url, params=params, **kwargs)
                response.raise_for_status()

                data = response.json()
                results.extend(data.get("result", []))

                next_page_id = data.get("next_page_id")
                if not next_page_id:
                    break

                params["_page_id"] = next_page_id

        except httpx.TimeoutException as exc:
            logging.error(f"Timeout error while fetching {url}: {exc}")
            raise WapiRequestException(exc) from exc
        except httpx.HTTPStatusError as exc:
            logging.error(f"HTTP status error while fetching {url}: {exc}")
            raise WapiRequestException(response.text) from exc
        except httpx.RequestError as exc:
            logging.error(f"Request error while fetching {url}: {exc}")
            raise WapiRequestException(str(exc)) from exc

        return results



import csv
import os
from logging import getLogger

from ibx_sdk.nios.csv.enums import ImportActionEnum

LOG = getLogger(__name__)


def extract_columns(item) -> list:
    """Extract column names from a single item."""
    return item.model_dump(
        by_alias=True, exclude_defaults=False, exclude_none=True
    ).keys()


def get_header(*, data: list) -> list:
    """Generate a unique header from the given data."""
    header_columns = []
    for item in data:
        cols = extract_columns(item)
        for col in cols:
            if col not in header_columns:
                header_columns.append(col)

    LOG.debug(header_columns)
    return header_columns


def output_to_file(
    *,
    filename: str,
    data: list,
    import_action: str = None,
    output_dir: str = None,
    file_prefix: str = None,
) -> None:
    """
    Generate a CSV file from the given data.

    Args:
        filename: csv filename or object name
        data: list of objects
        import_action: optional import-action to be added to the header
        output_dir: output to a specific directory
        file_prefix: optional file name prefix

    Returns:
        None
    """
    if filename.endswith(".csv"):
        output_file_name = filename
    else:
        output_file_name = f"{filename}.csv"

    if file_prefix:
        output_file_name = "-".join([file_prefix, output_file_name])

    if output_dir:
        output_file_name = os.path.join(output_dir, output_file_name)

    LOG.info(
        "Writing Infoblox NIOS %s data to CSV file %s",
        filename,
        output_file_name,
    )

    # if there's no data do not write to file
    if len(data) == 0:
        LOG.warning(
            "Skipping %s file, no data to write to file", output_file_name
        )
        return

    header = get_header(data=data)
    if import_action:
        LOG.debug("Adding import-action to header using %s", import_action)
        header.insert(1, "import-action")

    with open(output_file_name, "w") as f:
        mywriter = csv.DictWriter(f, fieldnames=header, extrasaction="ignore")
        mywriter.writeheader()
        for row in data:
            if import_action is not None:
                row.import_action = ImportActionEnum(import_action)
            mywriter.writerow(
                row.model_dump(
                    by_alias=True, exclude_defaults=False, exclude_none=True
                )
            )



"""Exceptions module for Infoblox Cloud SDK

Defines a hierarchy of custom exceptions for robust error handling
and clear semantics when interacting with the Infoblox Cloud API.
"""

__all__ = [
    "BaseCloudException",
    "ApiInvalidParameterException",
    "ApiRequestException",
]

class BaseCloudException(Exception):
    """
    Base exception for all Infoblox Cloud SDK errors.

    Inherit from this class to create SDK-specific exceptions.
    """
    pass

class ApiInvalidParameterException(BaseCloudException):
    """
    Raised when invalid arguments or parameters are passed to SDK methods.

    Indicates that the caller provided bad input that the SDK cannot
    resolve into a valid API call (e.g., unknown resource path).
    """
    def __init__(self, message: str):
        super().__init__(f"Invalid parameter: {message}")
        self.message = message

class ApiRequestException(BaseCloudException):
    """
    Raised when an HTTP request to the Infoblox Cloud API fails.

    Captures HTTP errors, non-2xx responses, or API-level error payloads.

    Attributes:
        status_code (int | None): HTTP status code of the failed response.
        payload (Any | None): Parsed JSON payload if available.
    """
    def __init__(
        self,
        message: str,
        status_code: int = None,
        payload: object = None,
        original_exception: Exception = None,
    ):
        """
        Initialize the ApiRequestException.

        Args:
            message: Human-readable error description.
            status_code: HTTP status code, if applicable.
            payload: Parsed JSON payload returned by the API.
            original_exception: Underlying exception instance, for chaining.
        """
        full_msg = "Request failed"
        if status_code is not None:
            full_msg += f" [{status_code}]"
        full_msg += f": {message}"
        super().__init__(full_msg)
        self.status_code = status_code
        self.payload = payload
        # Chain original exception for traceback clarity
        if original_exception is not None:
            self.__cause__ = original_exception



#!/usr/bin/env python3
"""
Copyright 2023 Infoblox

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import getpass
import sys

import click
from click_option_group import optgroup

from ibx_sdk.logger.ibx_logger import init_logger, increase_log_level
from ibx_sdk.nios.exceptions import WapiRequestException
from ibx_sdk.nios.gift import Gift

log = init_logger(
    logfile_name="wapi.log",
    logfile_mode="a",
    console_log=True,
    level="info",
    max_size=100000,
    num_logs=1,
)

wapi = Gift()

help_text = """
CSV Export by object
"""


@click.command(
    help=help_text,
    context_settings=dict(
        max_content_width=95, help_option_names=["-h", "--help"]
    ),
)
@optgroup.group("Required Parameters")
@optgroup.option(
    "-g", "--grid-mgr", required=True, help="Infoblox Grid Manager"
)
@optgroup.option("-f", "--filename", help="Infoblox WAPI CSV export file name")
@optgroup.group("Optional Parameters")
@optgroup.option(
    "-u",
    "--username",
    default="admin",
    show_default=True,
    help="Infoblox admin username",
)
@optgroup.option(
    "-w",
    "--wapi-ver",
    default="2.11",
    show_default=True,
    help="Infoblox WAPI version",
)
@optgroup.option(
    "-o", "--obj", default="network", help="WAPI export object type"
)
@optgroup.group("Logging Parameters")
@optgroup.option("--debug", is_flag=True, help="enable verbose debug output")
def main(
    grid_mgr: str,
    filename: str,
    username: str,
    wapi_ver: str,
    obj: str,
    debug: bool,
) -> None:
    """
    CSV Export

    Args:
        debug (bool): If True, it sets the log level to DEBUG. Default is False.
        grid_mgr (str): Manager for the wapi grid.
        username (str): Username for the wapi connection.
        wapi_ver (str): Version of wapi.
        obj (str): Object to be exported to CSV.
        filename (str): Filename/path where the CSV will be exported.

    Returns:
        None

    Raises:
        WapiRequestException: If unable to connect with the provided wapi parameters.
        SystemExit: The function exits the system upon completion or upon encounter of an error.

    """
    if debug:
        increase_log_level()

    wapi.grid_mgr = grid_mgr
    wapi.wapi_ver = wapi_ver
    password = getpass.getpass(f"Enter password for [{username}]: ")

    try:
        wapi.connect(username=username, password=password)
    except WapiRequestException as err:
        log.error(err)
        sys.exit(1)
    else:
        log.info("connected to Infoblox grid manager %s", wapi.grid_mgr)

    try:
        wapi.csv_export(wapi_object=obj, filename=filename)
    except WapiRequestException as err:
        log.error(err)
        sys.exit(1)

    sys.exit()


if __name__ == "__main__":
    main()



"""
Copyright 2023 Infoblox

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import os
import socket
from collections import Counter
from logging.handlers import RotatingFileHandler
from typing import Optional

import coloredlogs
from syslog_rfc5424_formatter import RFC5424Formatter


class CountingHandler(logging.Handler):
    def __init__(self):
        super().__init__()
        self.counts = Counter()

    def emit(self, record: logging.LogRecord) -> None:
        self.counts[record.levelname] += 1


log_levels = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}


def init_logger(
    logfile_name: str,
    logfile_mode: Optional[str] = "w",
    console_log: Optional[bool] = None,
    log_format: Optional[str] = None,
    max_size: Optional[int] = None,
    num_logs: Optional[int] = None,
    level: Optional[str] = None,
) -> logging.Logger:
    """
    Create and return a custom file/console logger.

    Args:
        logfile_name (str): The name of the log file.
        logfile_mode (str, optional): Specify the mode for the log file. 'a' for
            append or 'w' for write mode for basic file logging.
        console_log (bool, optional): If True, create a colored console log.
        log_format (str, optional): Specify the format for the log messages.
        max_size (int, optional): Specify the maximum size for the log file (in
            bytes) if you want to use a rotating log file handler instead of a
            standard log file handler.
        num_logs (int, optional): Specify the number of log files to keep if you
            want to use a rotating log file handler instead of a standard log
            file handler.
        level (str, optional): Specify a string value for the logging level. This
            field is case-insensitive and will default to 'INFO' if not provided.

    Returns:
        logging.Logger: The root logger.

    Example:

    **Basic FileHandler**
    ```python
    from ibx_sdk.logger.ibx_logger import init_logger

    # Initialize a basic logger to a file using write mode
    log = init_logger('logs/mylog.log', 'w')
    ```

    **Advanced RotatingFileHandler**
    ```python
    from ibx_sdk.logger.ibx_logger import init_logger

    # Initialize rotating file logging

    log = init_logger(
        logfile_name='logs/mylog.log',
        max_size=1000000,
        num_logs=10
    )
    ```
    """
    if level:
        log_level = log_levels.get(level.upper(), logging.INFO)
    else:
        log_level = logging.INFO

    if log_format:
        log_fmt = log_format
    else:
        log_fmt = (
            "%(asctime)s [%(filename)s:%(lineno)d - %(funcName)s()] %(levelname)s %(message)s"
        )

    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)

    # feature - logfile path
    if os.getenv("IBX_LOGDIR"):
        if not os.path.exists(os.getenv("IBX_LOGDIR")):
            try:
                os.makedirs(os.getenv("IBX_LOGDIR"))
            except FileExistsError as err:
                logging.warning(err)
            except OSError as err:
                logging.warning(err)
        logfile_name = os.path.join(os.getenv("IBX_LOGDIR"), logfile_name)

    if num_logs and max_size:
        lfh = RotatingFileHandler(
            filename=logfile_name, maxBytes=max_size, backupCount=num_logs
        )
    else:
        lfh = logging.FileHandler(
            filename=logfile_name,
            mode=logfile_mode,
        )
    lfh.setFormatter(logging.Formatter(log_fmt))
    lfh.setLevel(log_level)
    root_logger.addHandler(lfh)

    if console_log:
        init_console_logger(level, log_format)

    root_logger.debug("using logfile_name = %s", logfile_name)
    return root_logger


def init_remote_logger(
    address: tuple[str, int] = ("localhost", 5140),
    facility: str = "local0",
    protocol: str = "TCP",
    level: Optional[str] = None,
) -> logging.Logger:
    if level:
        log_level = log_levels.get(level.upper())
    else:
        log_level = logging.INFO
    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)
    socktype = socket.SOCK_STREAM if protocol == "TCP" else socket.SOCK_DGRAM
    handler = logging.handlers.SysLogHandler(
        address=address, facility=facility, socktype=socktype
    )
    handler.append_nul = False
    handler.setFormatter(RFC5424Formatter())
    root_logger.addHandler(handler)
    current_fmt = handler.formatter._fmt
    new_fmt = logging.Formatter(current_fmt + "\n")
    handler.setFormatter(new_fmt)
    return root_logger


def init_console_logger(
    level: Optional[str] = None, log_format: Optional[str] = None
):
    """
    Initialize a colored console logger with optional custom logging level.

    Args:
        level (str, optional): Specify a string value of the logging level.
            This field is case-insensitive and will default to 'INFO' if not provided.
        log_format (str, optional): Specify the format for the log messages.
            This field has a default value.


    Example:

    **Initialize Colored console logger with a custom logging level**
    ```python
    from ibx_sdk.logger.ibx_logger import init_console_logger

    # Initialize a console logger with a custom logging level
    init_console_logger(level='DEBUG')
    ```

    """
    if level:
        log_level = log_levels.get(level.upper(), logging.INFO)
    else:
        log_level = logging.INFO
    coloredlogs.DEFAULT_FIELD_STYLES["filename"] = dict(color="blue")

    if log_format:
        log_fmt = log_format
    else:
        log_fmt = (
            "%(asctime)s [%(filename)s:%(lineno)d] %(levelname)s %(message)s"
        )
    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)
    coloredlogs.install(logger=root_logger, level=log_level, fmt=log_fmt)
    return root_logger


def increase_log_level(handler_type: str = "both") -> None:
    """
    Increase the logging level of the root logger and specific handlers.

    Args:
        handler_type (str, optional): Specify the type of handlers to increase
            the logging level for. Can be 'both' (default), 'console', or 'file'.

    Returns:
        None

    Example:

    **Increase Log Level**
    ```python
    from your_module import increase_log_level

    # Increase the logging level for both console and file handlers
    increase_log_level()

    # Increase the logging level for console handlers only - use 'both' to increase the file and
    # console handlers
    increase_log_level(handler_type='console')
    ```
    """

    defined_levels = sorted(set(_defined_levels().values()))
    root_logger = logging.getLogger()
    current_level = root_logger.getEffectiveLevel()
    current_index = defined_levels.index(current_level)
    selected_index = max(0, current_index - 1)
    root_logger.setLevel(defined_levels[selected_index])
    for handle in root_logger.handlers:
        if isinstance(
            handle, coloredlogs.StandardErrorHandler
        ) and handler_type in [
            "both",
            "console",
        ]:
            current_index = defined_levels.index(handle.level)
            selected_index = max(0, current_index - 1)
            coloredlogs.set_level(defined_levels[selected_index])
        elif (
            isinstance(handle, logging.FileHandler)
            or isinstance(handle, logging.handlers.RotatingFileHandler)
            and handler_type in ["both", "file"]
        ):
            current_index = defined_levels.index(handle.level)
            selected_index = max(0, current_index - 1)
            handle.setLevel(defined_levels[selected_index])


def set_log_level(level: str, handler_type: str = "both") -> None:
    """
    Set the logging level for the root logger and specific handlers.

    Args:
        level (str): The desired logging level, e.g., 'DEBUG', 'INFO', 'WARNING'.
            This field is case-insensitive.
        handler_type (str, optional): Specify the type of handlers to set the
            logging level for. Can be 'both' (default), 'console', or 'file'.

    Returns:
        None

    Example:

    **Set the logging level for both console and file handlers to 'DEBUG'**

    ```python
    from your_module import set_log_level

    # Set the logging level for both console and file handlers
    set_log_level(level='DEBUG')

    # or more explicitly using:
    set_log_level(level='DEBUG', handler_type='both')

    # Set the logging level for console handlers only to 'INFO'
    set_log_level(level='INFO', handler_type='console')
    ```

    """
    log_level = log_levels.get(level.upper(), logging.INFO)
    root_logger = logging.getLogger()
    if handler_type == "both":
        root_logger.setLevel(log_level)
    for handle in root_logger.handlers:
        if isinstance(
            handle, coloredlogs.StandardErrorHandler
        ) and handler_type in [
            "both",
            "console",
        ]:
            coloredlogs.set_level(log_level)
        elif (
            isinstance(handle, logging.FileHandler)
            or isinstance(handle, logging.handlers.RotatingFileHandler)
            and handler_type in ["both", "file"]
        ):
            handle.setLevel(log_level)


def _defined_levels():
    defined_levels = {}
    for name in dir(logging):
        if name.isupper():
            value = getattr(logging, name)
            if isinstance(value, int):
                defined_levels[name] = value
    return defined_levels



class BaseWapiException(Exception):
    """BaseWapiException class"""


class WapiInvalidParameterException(BaseWapiException):
    """WapiInvalidParameterException class - raised when invalid args/params passed to methods"""


class WapiRequestException(BaseWapiException):
    """WapiRequestException class - returns error(s) returned from Infoblox WAPI calls"""

    def __init__(self, msg):
        super().__init__(f"wapi error - {msg}")
